import os
from gluoncv import data

class COCODetectionTiny(data.COCODetection):
    CLASSES = ['bicycle', 'motorcycle']

    def __init__(self, root=os.path.join('~', '.mxnet', 'datasets', 'tiny_coco'),
                 splits=('instances_val2017_tiny',), **kwargs):
        super().__init__(root=root, splits=splits, **kwargs)

class COCOInstanceTiny(data.COCOInstance):
    CLASSES = ['bicycle', 'motorcycle']

    def __init__(self, root=os.path.join('~', '.mxnet', 'datasets', 'tiny_coco'),
                 splits=('instances_val2017_tiny',), **kwargs):
        super().__init__(root=root, splits=splits, **kwargs)

class VOCDetectionTiny(data.VOCDetection):
    CLASSES = ['motorbike', 'person']

    def __init__(self, root=os.path.join('~', '.mxnet', 'datasets', 'tiny_motorbike'),
                 splits=(('tiny_motorbike', 'trainval'),), **kwargs):
        super().__init__(root=root, splits=splits, **kwargs)

class VOCSegmentationTiny(data.VOCSegmentation):
    CLASSES = ['motorbike', 'person']
    BASE_DIR = 'tiny_motorbike'

    def __init__(self, root=os.path.expanduser(os.path.join('~', '.mxnet', 'datasets', 'tiny_motorbike')),
                 split='train', **kwargs):
        super().__init__(root=root, split=split, **kwargs)



from __future__ import print_function
from __future__ import division

import os.path as osp
import mxnet as mx
import numpy as np
import gluoncv as gcv

from mxnet import autograd, gluon
from math import pi, cos
from gluoncv.utils import LRScheduler, LRSequential

def compare(obj, niter, expect, **kwargs):
    np.testing.assert_allclose(expect, obj.__call__(niter), **kwargs)

def test_sanity():
    N = 1000
    constant = LRScheduler('constant', base_lr=0, target_lr=1, niters=N)
    linear = LRScheduler('linear', base_lr=1, target_lr=2, niters=N)
    cosine = LRScheduler('cosine', base_lr=3, target_lr=1, niters=N)
    poly = LRScheduler('poly', base_lr=1, target_lr=0, niters=N, power=2)
    step = LRScheduler('step', base_lr=1, target_lr=0, niters=N,
                       step_iter=[100, 500], step_factor=0.1)

    compare(constant, 0, 0)
    compare(constant, N-1, 0)
    compare(linear, 0, 1)
    compare(linear, N-1, 2)
    compare(cosine, 0, 3)
    compare(cosine, N-1, 1)
    compare(poly, 0, 1)
    compare(poly, N-1, 0)
    compare(step, 0, 1)
    compare(step, 100, 0.1)
    compare(step, 500, 0.01)
    compare(step, N-1, 0.01)

def test_single_method():
    N = 1000
    constant = LRScheduler('constant', base_lr=0, target_lr=1, niters=N)
    linear = LRScheduler('linear', base_lr=1, target_lr=2, niters=N)
    cosine = LRScheduler('cosine', base_lr=3, target_lr=1, niters=N)
    poly = LRScheduler('poly', base_lr=1, target_lr=0, niters=N, power=2)
    step = LRScheduler('step', base_lr=1, target_lr=0, niters=N,
                       step_iter=[100, 500], step_factor=0.1)
    step2 = LRScheduler('step', base_lr=1, target_lr=0,
                        nepochs=2, iters_per_epoch=N/2,
                        step_iter=[100, 500], step_factor=0.1)
    step3 = LRScheduler('step', base_lr=1, target_lr=0,
                        nepochs=100, iters_per_epoch=N/100,
                        step_epoch=[10, 50], step_factor=0.1)

    # Test numerical value
    for i in range(N):
        compare(constant, i, 0)

        expect_linear = 2 + (1 - 2) * (1 - i / (N - 1))
        compare(linear, i, expect_linear)

        expect_cosine = 1 + (3 - 1) * ((1 + cos(pi * i / (N-1))) / 2)
        compare(cosine, i, expect_cosine)

        expect_poly = 0 + (1 - 0) * (pow(1 - i / (N-1), 2))
        compare(poly, i, expect_poly)

        if i < 100:
            expect_step = 1
        elif i < 500:
            expect_step = 0.1
        else:
            expect_step = 0.01
        compare(step, i, expect_step)
        compare(step2, i, expect_step)
        compare(step3, i, expect_step)

    # Test out-of-range updates
    for i in range(10):
        constant.update(i - 3)
        linear.update(i - 3)
        cosine.update(i - 3)
        poly.update(i - 3)

def test_composed_method():
    N = 1000
    constant = LRScheduler('constant', base_lr=0, target_lr=1, niters=N)
    linear = LRScheduler('linear', base_lr=1, target_lr=2, niters=N)
    cosine = LRScheduler('cosine', base_lr=3, target_lr=1, niters=N)
    poly = LRScheduler('poly', base_lr=1, target_lr=0, niters=N, power=2)
    # components with niters=0 will be ignored
    null_cosine = LRScheduler('cosine', base_lr=3, target_lr=1, niters=0)
    null_poly = LRScheduler('cosine', base_lr=3, target_lr=1, niters=0)
    step = LRScheduler('step', base_lr=1, target_lr=0, niters=N,
                       step_iter=[100, 500], step_factor=0.1)
    arr = LRSequential([constant, null_cosine, linear, cosine, null_poly, poly, step])
    # constant
    for i in range(N):
        compare(arr, i, 0)
    # linear
    for i in range(N, 2*N):
        expect_linear = 2 + (1 - 2) * (1 - (i - N) / (N - 1))
        compare(arr, i, expect_linear)
    # cosine
    for i in range(2*N, 3*N):
        expect_cosine = 1 + (3 - 1) * ((1 + cos(pi * (i - 2*N) / (N - 1))) / 2)
        compare(arr, i, expect_cosine)
    # poly
    for i in range(3*N, 4*N):
        expect_poly = 0 + (1 - 0) * (pow(1 - (i - 3*N) / (N - 1), 2))
        compare(arr, i, expect_poly)
    for i in range(4*N, 5*N):
        if i - 4*N < 100:
            expect_step = 1
        elif i - 4*N < 500:
            expect_step = 0.1
        else:
            expect_step = 0.01
        compare(arr, i, expect_step)
    # out-of-bound index
    compare(arr, 10*N, 0.01)
    compare(arr, -1, 0)

def test_params():
    N = 1000
    linear = LRScheduler('linear', base_lr=1, target_lr=2, niters=N)
    linear2 = LRScheduler('linear', baselr=1, targetlr=2, niters=N)
    linear3 = LRScheduler('linear', base_lr=1, target_lr=2, niters=N,
                          baselr=0, targetlr=1)

    linear4 = LRScheduler('linear', base_lr=1, target_lr=2, niters=N/2)
    linear5 = LRScheduler('linear', base_lr=1, target_lr=2, niters=N/2,
                          nepochs=N/2, iters_per_epoch=2)

    compare(linear, 0, 1)
    compare(linear, N-1, 2)
    compare(linear2, 0, 1)
    compare(linear2, N-1, 2)
    compare(linear3, 0, 1)
    compare(linear3, N-1, 2)
    compare(linear4, 0, 1)
    compare(linear4, N/2-1, 2)
    compare(linear5, 0, 1)
    compare(linear5, N/2-1, 1.5, rtol=0.01)
    compare(linear5, N-1, 2)

if __name__ == '__main__':
    import nose
    nose.runmodule()



"""Visualize image."""
import numpy as np
import mxnet as mx

def plot_image(img, ax=None, reverse_rgb=False):
    """Visualize image.

    Parameters
    ----------
    img : numpy.ndarray or mxnet.nd.NDArray
        Image with shape `H, W, 3`.
    ax : matplotlib axes, optional
        You can reuse previous axes if provided.
    reverse_rgb : bool, optional
        Reverse RGB<->BGR orders if `True`.

    Returns
    -------
    matplotlib axes
        The ploted axes.

    Examples
    --------

    from matplotlib import pyplot as plt
    ax = plot_image(img)
    plt.show()
    """
    from matplotlib import pyplot as plt
    if ax is None:
        # create new axes
        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1)
    if isinstance(img, mx.nd.NDArray):
        img = img.asnumpy()
    img = img.copy()
    if reverse_rgb:
        img[:, :, (0, 1, 2)] = img[:, :, (2, 1, 0)]
    ax.imshow(img.astype(np.uint8))
    return ax

def cv_plot_image(img, scale=1, upperleft_txt=None, upperleft_txt_corner=(10, 100),
                  left_txt_list=None, left_txt_corner=(10, 150),
                  title_txt_list=None, title_txt_corner=(500, 50),
                  canvas_name='demo'):
    """Visualize image with OpenCV.

    Parameters
    ----------
    img : numpy.ndarray or mxnet.nd.NDArray
        Image with shape `H, W, 3`.
    scale : float
        The scaling factor of the output image
    upperleft_txt : str, optional, default is None
        If presents, will print the string at the upperleft corner
    upperleft_txt_corner : tuple, optional, default is (10, 100)
        The bottomleft corner of `upperleft_txt`
    left_txt_list : list of str, optional, default is None
        If presents, will print each string in the list close to the left
    left_txt_corner : tuple, optional, default is (10, 150)
        The bottomleft corner of `left_txt_list`
    title_txt_list : list of str, optional, default is None
        If presents, will print each string in the list close to the top
    title_txt_corner : tuple, optional, default is (500, 50)
        The bottomleft corner of `title_txt_list`
    canvas_name : str, optional, default is 'demo'
        The name of the canvas to plot the image

    Examples
    --------

    from matplotlib import pyplot as plt
    ax = plot_image(img)
    plt.show()
    """
    from ..filesystem import try_import_cv2
    cv2 = try_import_cv2()

    if isinstance(img, mx.nd.NDArray):
        img = img.asnumpy()

    height, width, _ = img.shape
    img = cv2.resize(img, (int(width * scale), int(height * scale)))
    if upperleft_txt is not None:
        font = cv2.FONT_HERSHEY_SIMPLEX
        bottomLeftCornerOfText = upperleft_txt_corner
        fontScale = 1
        fontColor = (255, 255, 255)
        thickness = 3

        cv2.putText(img, upperleft_txt, bottomLeftCornerOfText,
                    font, fontScale, fontColor, thickness)

    if left_txt_list is not None:
        starty = left_txt_corner[1]
        for txt in left_txt_list:
            font = cv2.FONT_HERSHEY_SIMPLEX
            bottomLeftCornerOfText = (left_txt_corner[0], starty)
            fontScale = 1
            fontColor = (255, 255, 255)
            thickness = 1

            cv2.putText(img, txt, bottomLeftCornerOfText,
                        font, fontScale, fontColor, thickness)

            starty += 30

    if title_txt_list is not None:
        font = cv2.FONT_HERSHEY_SIMPLEX
        bottomLeftCornerOfText = title_txt_corner
        fontScale = 1
        fontColor = (255, 255, 255)
        thickness = 3

        for txt in title_txt_list:
            cv2.putText(img, txt, bottomLeftCornerOfText,
                        font, fontScale, fontColor, thickness)
            bottomLeftCornerOfText = (bottomLeftCornerOfText[0] + 100,
                                      bottomLeftCornerOfText[1] + 50)

    canvas = cv2.cvtColor(img, cv2.COLOR_RGB2BGR)
    cv2.imshow(canvas_name, canvas)



# pylint: disable=missing-function-docstring, missing-class-docstring, unused-argument
"""R2Plus1D, https://arxiv.org/abs/1711.11248. Code adapted from
https://github.com/pytorch/vision/blob/master/torchvision/models/video/resnet.py."""
import torch
import torch.nn as nn
from torch.nn import BatchNorm3d

__all__ = ['R2Plus1D', 'r2plus1d_v1_resnet18_kinetics400', 'r2plus1d_v1_resnet34_kinetics400',
           'r2plus1d_v1_resnet50_kinetics400', 'r2plus1d_v1_resnet101_kinetics400',
           'r2plus1d_v1_resnet152_kinetics400', 'r2plus1d_v1_resnet50_custom']


def conv3x1x1(in_planes, out_planes, spatial_stride=1, temporal_stride=1, dilation=1):
    """3x1x1 convolution with padding"""
    return nn.Conv3d(in_channels=in_planes,
                     out_channels=out_planes,
                     kernel_size=(3, 1, 1),
                     stride=(temporal_stride, spatial_stride, spatial_stride),
                     padding=(dilation, 0, 0),
                     dilation=dilation,
                     bias=False)


class Conv2Plus1D(nn.Module):

    def __init__(self,
                 inplanes,
                 planes,
                 midplanes,
                 stride=1,
                 padding=1,
                 norm_layer=BatchNorm3d,
                 norm_kwargs=None,
                 **kwargs):
        super(Conv2Plus1D, self).__init__()

        self.conv1 = nn.Conv3d(in_channels=inplanes,
                               out_channels=midplanes,
                               kernel_size=(1, 3, 3),
                               stride=(1, stride, stride),
                               padding=(0, padding, padding),
                               bias=False)
        self.bn1 = norm_layer(num_features=midplanes,
                              **({} if norm_kwargs is None else norm_kwargs))
        self.relu = nn.ReLU(inplace=False)
        self.conv2 = nn.Conv3d(in_channels=midplanes,
                               out_channels=planes,
                               kernel_size=(3, 1, 1),
                               stride=(stride, 1, 1),
                               padding=(padding, 0, 0),
                               bias=False)

    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.conv2(x)
        return x


class BasicBlock(nn.Module):
    expansion = 1

    def __init__(self, inplanes, planes, stride=1, downsample=None,
                 norm_layer=BatchNorm3d, norm_kwargs=None, layer_name='',
                 **kwargs):
        super(BasicBlock, self).__init__()
        self.downsample = downsample

        midplanes = (inplanes * planes * 3 * 3 * 3) // (inplanes * 3 * 3 + 3 * planes)
        self.conv1 = Conv2Plus1D(inplanes, planes, midplanes, stride)
        self.bn1 = norm_layer(num_features=planes,
                              **({} if norm_kwargs is None else norm_kwargs))
        self.relu = nn.ReLU(inplace=False)
        self.conv2 = Conv2Plus1D(planes, planes, midplanes)
        self.bn2 = norm_layer(num_features=planes,
                              **({} if norm_kwargs is None else norm_kwargs))

    def forward(self, x):
        identity = x
        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out = out + identity
        out = self.relu(out)
        return out


class Bottleneck(nn.Module):
    expansion = 4

    def __init__(self, inplanes, planes, stride=1, downsample=None,
                 norm_layer=BatchNorm3d, norm_kwargs=None, layer_name='',
                 **kwargs):
        super(Bottleneck, self).__init__()
        self.downsample = downsample

        midplanes = (inplanes * planes * 3 * 3 * 3) // (inplanes * 3 * 3 + 3 * planes)

        # 1x1x1
        self.conv1 = nn.Conv3d(in_channels=inplanes, out_channels=planes, kernel_size=1, bias=False)
        self.bn1 = norm_layer(num_features=planes,
                              **({} if norm_kwargs is None else norm_kwargs))
        self.relu = nn.ReLU(inplace=False)

        # Second kernel
        self.conv2 = Conv2Plus1D(planes, planes, midplanes, stride)
        self.bn2 = norm_layer(num_features=planes,
                              **({} if norm_kwargs is None else norm_kwargs))

        self.conv3 = nn.Conv3d(in_channels=planes, out_channels=planes * self.expansion,
                               kernel_size=1, bias=False)
        self.bn3 = norm_layer(num_features=planes * self.expansion,
                              **({} if norm_kwargs is None else norm_kwargs))

    def forward(self, x):
        identity = x
        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out = out + identity
        out = self.relu(out)
        return out


class R2Plus1D(nn.Module):
    r"""The R2+1D network.
    A Closer Look at Spatiotemporal Convolutions for Action Recognition.
    CVPR, 2018. https://arxiv.org/abs/1711.11248
    """
    def __init__(self, num_classes, block, layers, dropout_ratio=0.5,
                 num_segment=1, num_crop=1, feat_ext=False,
                 init_std=0.001, partial_bn=False,
                 norm_layer=BatchNorm3d, norm_kwargs=None, **kwargs):
        super(R2Plus1D, self).__init__()
        self.partial_bn = partial_bn
        self.dropout_ratio = dropout_ratio
        self.init_std = init_std
        self.num_segment = num_segment
        self.num_crop = num_crop
        self.feat_ext = feat_ext
        self.inplanes = 64
        self.feat_dim = 512 * block.expansion

        self.conv1 = nn.Conv3d(in_channels=3, out_channels=45, kernel_size=(1, 7, 7),
                               stride=(1, 2, 2), padding=(0, 3, 3), bias=False)
        self.bn1 = norm_layer(num_features=45, **({} if norm_kwargs is None else norm_kwargs))
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = conv3x1x1(in_planes=45, out_planes=64)
        self.bn2 = norm_layer(num_features=64, **({} if norm_kwargs is None else norm_kwargs))

        if self.partial_bn:
            if norm_kwargs is not None:
                norm_kwargs['use_global_stats'] = True
            else:
                norm_kwargs = {}
                norm_kwargs['use_global_stats'] = True

        self.layer1 = self._make_res_layer(block=block,
                                           planes=64,
                                           blocks=layers[0],
                                           layer_name='layer1_')
        self.layer2 = self._make_res_layer(block=block,
                                           planes=128,
                                           blocks=layers[1],
                                           stride=2,
                                           layer_name='layer2_')
        self.layer3 = self._make_res_layer(block=block,
                                           planes=256,
                                           blocks=layers[2],
                                           stride=2,
                                           layer_name='layer3_')
        self.layer4 = self._make_res_layer(block=block,
                                           planes=512,
                                           blocks=layers[3],
                                           stride=2,
                                           layer_name='layer4_')

        self.avgpool = nn.AdaptiveAvgPool3d(output_size=(1, 1, 1))
        self.dropout = nn.Dropout(self.dropout_ratio)
        self.fc = nn.Linear(in_features=self.feat_dim, out_features=num_classes)
        nn.init.normal_(self.fc.weight, 0, self.init_std)
        nn.init.constant_(self.fc.bias, 0)

    def _make_res_layer(self,
                        block,
                        planes,
                        blocks,
                        stride=1,
                        norm_layer=BatchNorm3d,
                        norm_kwargs=None,
                        layer_name=''):
        """Build each stage of a ResNet"""
        downsample = None

        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                nn.Conv3d(in_channels=self.inplanes,
                          out_channels=planes * block.expansion,
                          kernel_size=1,
                          stride=(stride, stride, stride),
                          bias=False),
                norm_layer(num_features=planes * block.expansion,
                           **({} if norm_kwargs is None else norm_kwargs)))

        layers = []
        layers.append(block(inplanes=self.inplanes,
                            planes=planes,
                            stride=stride,
                            downsample=downsample))

        self.inplanes = planes * block.expansion
        for _ in range(1, blocks):
            layers.append(block(inplanes=self.inplanes, planes=planes))

        return nn.Sequential(*layers)

    def forward(self, x):
        bs, _, _, _, _ = x.size()
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.conv2(x)
        x = self.bn2(x)
        x = self.relu(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)

        x = self.avgpool(x)
        x = x.view(bs, -1)

        if self.feat_ext:
            return x

        x = self.fc(self.dropout(x))
        return x


def r2plus1d_v1_resnet18_kinetics400(cfg):
    model = R2Plus1D(num_classes=cfg.CONFIG.DATA.NUM_CLASSES,
                     block=BasicBlock,
                     layers=[2, 2, 2, 2],
                     num_segment=cfg.CONFIG.DATA.NUM_SEGMENT,
                     num_crop=cfg.CONFIG.DATA.NUM_CROP,
                     feat_ext=cfg.CONFIG.INFERENCE.FEAT,
                     partial_bn=cfg.CONFIG.MODEL.PARTIAL_BN)
    if cfg.CONFIG.MODEL.PRETRAINED:
        from ..model_store import get_model_file
        model.load_state_dict(torch.load(get_model_file('r2plus1d_v1_resnet18_kinetics400',
                                                        tag=cfg.CONFIG.MODEL.PRETRAINED)))
    return model


def r2plus1d_v1_resnet34_kinetics400(cfg):
    model = R2Plus1D(num_classes=cfg.CONFIG.DATA.NUM_CLASSES,
                     block=BasicBlock,
                     layers=[3, 4, 6, 3],
                     num_segment=cfg.CONFIG.DATA.NUM_SEGMENT,
                     num_crop=cfg.CONFIG.DATA.NUM_CROP,
                     feat_ext=cfg.CONFIG.INFERENCE.FEAT,
                     partial_bn=cfg.CONFIG.MODEL.PARTIAL_BN)

    if cfg.CONFIG.MODEL.PRETRAINED:
        from ..model_store import get_model_file
        model.load_state_dict(torch.load(get_model_file('r2plus1d_v1_resnet34_kinetics400',
                                                        tag=cfg.CONFIG.MODEL.PRETRAINED)))
    return model


def r2plus1d_v1_resnet50_kinetics400(cfg):
    model = R2Plus1D(num_classes=cfg.CONFIG.DATA.NUM_CLASSES,
                     block=Bottleneck,
                     layers=[3, 4, 6, 3],
                     num_segment=cfg.CONFIG.DATA.NUM_SEGMENT,
                     num_crop=cfg.CONFIG.DATA.NUM_CROP,
                     feat_ext=cfg.CONFIG.INFERENCE.FEAT,
                     partial_bn=cfg.CONFIG.MODEL.PARTIAL_BN)

    if cfg.CONFIG.MODEL.PRETRAINED:
        from ..model_store import get_model_file
        model.load_state_dict(torch.load(get_model_file('r2plus1d_v1_resnet50_kinetics400',
                                                        tag=cfg.CONFIG.MODEL.PRETRAINED)))
    return model


def r2plus1d_v1_resnet101_kinetics400(cfg):
    model = R2Plus1D(num_classes=cfg.CONFIG.DATA.NUM_CLASSES,
                     block=Bottleneck,
                     layers=[3, 4, 23, 3],
                     num_segment=cfg.CONFIG.DATA.NUM_SEGMENT,
                     num_crop=cfg.CONFIG.DATA.NUM_CROP,
                     feat_ext=cfg.CONFIG.INFERENCE.FEAT,
                     partial_bn=cfg.CONFIG.MODEL.PARTIAL_BN)

    if cfg.CONFIG.MODEL.PRETRAINED:
        from ..model_store import get_model_file
        model.load_state_dict(torch.load(get_model_file('r2plus1d_v1_resnet101_kinetics400',
                                                        tag=cfg.CONFIG.MODEL.PRETRAINED)))
    return model

def r2plus1d_v1_resnet152_kinetics400(cfg):
    model = R2Plus1D(num_classes=cfg.CONFIG.DATA.NUM_CLASSES,
                     block=Bottleneck,
                     layers=[3, 8, 36, 3],
                     num_segment=cfg.CONFIG.DATA.NUM_SEGMENT,
                     num_crop=cfg.CONFIG.DATA.NUM_CROP,
                     feat_ext=cfg.CONFIG.INFERENCE.FEAT,
                     partial_bn=cfg.CONFIG.MODEL.PARTIAL_BN)

    if cfg.CONFIG.MODEL.PRETRAINED:
        from ..model_store import get_model_file
        model.load_state_dict(torch.load(get_model_file('r2plus1d_v1_resnet152_kinetics400',
                                                        tag=cfg.CONFIG.MODEL.PRETRAINED)))
    return model


def r2plus1d_v1_resnet50_custom(cfg):
    model = R2Plus1D(num_classes=cfg.CONFIG.DATA.NUM_CLASSES,
                     block=Bottleneck,
                     layers=[3, 4, 6, 3],
                     num_segment=cfg.CONFIG.DATA.NUM_SEGMENT,
                     num_crop=cfg.CONFIG.DATA.NUM_CROP,
                     feat_ext=cfg.CONFIG.INFERENCE.FEAT,
                     partial_bn=cfg.CONFIG.MODEL.PARTIAL_BN)

    if cfg.CONFIG.MODEL.PRETRAINED:
        from ..model_store import get_model_file
        state_dict = torch.load(get_model_file('r2plus1d_v1_resnet50_kinetics400', tag=cfg.CONFIG.MODEL.PRETRAINED))
        for k in list(state_dict.keys()):
            # retain only backbone up to before the classification layer
            if k.startswith('fc'):
                del state_dict[k]

        msg = model.load_state_dict(state_dict, strict=False)
        assert set(msg.missing_keys) == {'fc.weight', 'fc.bias'}
        print("=> initialized from a R2+1D model pretrained on Kinetcis400 dataset")
    return model



"""KITTI Dataset. (KITTI Raw, KITTI Odom, KITTI Depth)
Vision meets Robotics: The KITTI Dataset, IJRR 2013
http://www.cvlibs.net/datasets/kitti/raw_data.php
Code partially borrowed from
https://github.com/nianticlabs/monodepth2/blob/master/datasets/kitti_dataset.py
"""
# pylint: disable=abstract-method, unused-import
# Copyright Niantic 2019. Patent Pending. All rights reserved.
#
# This software is licensed under the terms of the Monodepth2 licence
# which allows for non-commercial use only, the full terms of which are made
# available in the LICENSE file.

from __future__ import absolute_import, division, print_function

import os
import numpy as np
import PIL.Image as pil

from ...utils.filesystem import try_import_skimage
from .kitti_utils import generate_depth_map
from .mono_dataset import MonoDataset


class KITTIDataset(MonoDataset):
    """Superclass for different types of KITTI dataset loaders
    """

    def __init__(self, *args, **kwargs):
        super(KITTIDataset, self).__init__(*args, **kwargs)

        self.K = np.array([[0.58, 0, 0.5, 0],
                           [0, 1.92, 0.5, 0],
                           [0, 0, 1, 0],
                           [0, 0, 0, 1]], dtype=np.float32)

        self.full_res_shape = (1242, 375)
        self.side_map = {"2": 2, "3": 3,
                         "l": 2, "r": 3}

    def check_depth(self):
        line = self.filenames[0].split()
        scene_name = line[0]
        frame_index = int(line[1])

        velo_filename = os.path.join(
            self.data_path,
            scene_name,
            "velodyne_points/data/{:010d}.bin".format(int(frame_index)))

        return os.path.isfile(velo_filename)

    def get_color(self, folder, frame_index, side, do_flip):
        color = self.loader(self.get_image_path(folder, frame_index, side))

        if do_flip:
            color = color.transpose(pil.FLIP_LEFT_RIGHT)

        return color


class KITTIRAWDataset(KITTIDataset):
    """KITTI Raw Dataset.
    Parameters
    ----------
    data_path : string
        Path to KITTI RAW dataset folder. Default is '$(HOME)/.mxnet/datasets/kitti/kitti_data'

    Examples
    --------
    >>> from gluoncv.data.kitti.kitti_utils import dict_batchify_fn, readlines
    >>> train_filenames = os.path.join(
    >>>     os.path.expanduser("~"), '/.mxnet/datasets/kitti/splits/eigen_full/train_files.txt')
    >>> train_filenames = readlines(train_filenames)
    >>> # Create Dataset
    >>> trainset = gluoncv.data.KITTIRAWDataset(
    >>>         filenames=train_filenames, height=192, width=640,
    >>>         frame_idxs=[0], num_scales=4, is_train=True, img_ext='.png')
    >>> # Create Training Loader
    >>> train_data = gluon.data.DataLoader(
    >>>     trainset, batch_size=12, shuffle=True,
    >>>     batchify_fn=dict_batchify_fn, num_workers=12,
    >>>     pin_memory=True, last_batch='discard')
    """

    # pylint: disable=keyword-arg-before-vararg
    def __init__(self, data_path=os.path.join(
            os.path.expanduser("~"), '.mxnet/datasets/kitti/kitti_data'), *args, **kwargs):
        super(KITTIRAWDataset, self).__init__(data_path, *args, **kwargs)

    def get_image_path(self, folder, frame_index, side):
        f_str = "{:010d}{}".format(frame_index, self.img_ext)
        image_path = os.path.join(
            self.data_path, folder, "image_0{}/data".format(self.side_map[side]), f_str)
        return image_path

    def get_depth(self, folder, frame_index, side, do_flip):
        calib_path = os.path.join(self.data_path, folder.split("/")[0])

        velo_filename = os.path.join(
            self.data_path,
            folder,
            "velodyne_points/data/{:010d}.bin".format(int(frame_index)))

        depth_gt = generate_depth_map(calib_path, velo_filename, self.side_map[side])
        skimage = try_import_skimage()
        from skimage import transform
        depth_gt = skimage.transform.resize(
            depth_gt, self.full_res_shape[::-1], order=0, preserve_range=True, mode='constant')

        if do_flip:
            depth_gt = np.fliplr(depth_gt)

        return depth_gt


class KITTIOdomDataset(KITTIDataset):
    """KITTI dataset for odometry training and testing
    """

    # pylint: disable=keyword-arg-before-vararg
    def __init__(self, data_path=os.path.join(
            os.path.expanduser("~"), '.mxnet/datasets/kitti/kitti_odom'), *args, **kwargs):
        super(KITTIOdomDataset, self).__init__(data_path, *args, **kwargs)

    def get_image_path(self, folder, frame_index, side):
        f_str = "{:06d}{}".format(frame_index, self.img_ext)
        image_path = os.path.join(
            self.data_path,
            "sequences/{:02d}".format(int(folder)),
            "image_{}".format(self.side_map[side]),
            f_str)
        return image_path


class KITTIDepthDataset(KITTIDataset):
    """KITTI dataset which uses the updated ground truth depth maps
    """

    def __init__(self, *args, **kwargs):
        super(KITTIDepthDataset, self).__init__(*args, **kwargs)

    def get_image_path(self, folder, frame_index, side):
        f_str = "{:010d}{}".format(frame_index, self.img_ext)
        image_path = os.path.join(
            self.data_path,
            folder,
            "image_0{}/data".format(self.side_map[side]),
            f_str)
        return image_path

    def get_depth(self, folder, frame_index, side, do_flip):
        f_str = "{:010d}.png".format(frame_index)
        depth_path = os.path.join(
            self.data_path,
            folder,
            "proj_depth/groundtruth/image_0{}".format(self.side_map[side]),
            f_str)

        depth_gt = pil.open(depth_path)
        depth_gt = depth_gt.resize(self.full_res_shape, pil.NEAREST)
        depth_gt = np.array(depth_gt).astype(np.float32) / 256

        if do_flip:
            depth_gt = np.fliplr(depth_gt)

        return depth_gt



"""Mask RCNN default config"""
# pylint: disable=unused-variable,missing-function-docstring,bad-whitespace,missing-class-docstring
# from typing import Union, Tuple
from autocfg import dataclass, field

@dataclass
class MaskRCNN:
    # Backbone network.
    backbone = 'resnet50_v1b'  # base feature network
    # Final R-CNN non-maximum suppression threshold. You can specify < 0 or > 1 to disable NMS.
    nms_thresh = 0.5
    # Apply R-CNN NMS to top k detection results, use -1 to disable so that every Detection
    # result is used in NMS.
    nms_topk = -1
    # Only return top `post_nms` detection results, the rest is discarded.
    # Set to -1 to return all detections.
    post_nms = -1
    # ROI pooling mode. Currently support 'pool' and 'align'.
    roi_mode = 'align'
    # (height, width) of the ROI region.
    roi_size = (7, 7)
    # Feature map stride with respect to original image.
    # This is usually the ratio between original image size and feature map size.
    # For FPN, use a tuple of ints.
    strides = (4, 8, 16, 32, 64)
    # Clip bounding box prediction to to prevent exponentiation from overflowing.
    clip = 4.14

    # Anchors generation
    # ------------------
    # The width(and height) of reference anchor box.
    anchor_base_size = 16
    # The areas of anchor boxes.
    # We use the following form to compute the shapes of anchors:
    # .. math::
    # width_{anchor} = size_{base} \times scale \times \sqrt{ 1 / ratio}
    # height_{anchor} = size_{base} \times scale \times \sqrt{ratio}
    anchor_aspect_ratio = (0.5, 1, 2)
    # The aspect ratios of anchor boxes. We expect it to be a list or tuple.
    anchor_scales = (2, 4, 8, 16, 32)

    # Allocate size for the anchor boxes as (H, W).
    # Usually we generate enough anchors for large feature map, e.g. 128x128.
    # Later in inference we can have variable input sizes,
    # at which time we can crop corresponding anchors from this large
    # anchor map so we can skip re-generating anchors for each input.
    anchor_alloc_size = (384, 384)

    # number of channels used in RPN convolutional layers.
    rpn_channel = 256
    # IOU threshold for NMS. It is used to remove overlapping proposals.
    rpn_nms_thresh = 0.7
    # Maximum ground-truth number for each example. This is only an upper bound, not
    # necessarily very precise. However, using a very big number may impact the training speed.
    max_num_gt = 100
    # Gluon normalization layer to use. Default is none which will use frozen
    # batch normalization layer.
    norm_layer = None

    # FPN Options
    # -----------
    # Whether to use FPN.
    use_fpn = True
    # Number of filters for FPN output layers.
    num_fpn_filters = 256

    # Number of convolution layers to use in box head if batch normalization is not frozen.
    num_box_head_conv = 4
    # Number of filters for convolution layers in box head.
    # Only applicable if batch normalization is not frozen.
    num_box_head_conv_filters = 256
    # Number of hidden units for the last fully connected layer in box head.
    num_box_head_dense_filters = 1024

    # Input image short side size.
    image_short = 800
    # Maximum size of input image long side.
    image_max_size = 1333

    # Whether to enable custom model.
    custom_model = True
    # Whether to use automatic mixed precision
    amp = False
    # Whether to allocate memory statically.
    static_alloc = False

    # Ratio of mask output roi / input roi.
    # For model with FPN, this is typically 2.
    target_roi_scale = 2
    # Number of convolution blocks before deconv layer for mask head.
    # For FPN network this is typically 4.
    num_mask_head_convs = 4


@dataclass
class TrainCfg:
    # Whether load the imagenet pre-trained base
    pretrained_base = True
    # Batch size during training
    batch_size = 1
    # starting epoch
    start_epoch = 0
    # total epoch for training
    epochs = 26

    # Solver
    # ------
    # Learning rate.
    lr = 0.01
    # Decay rate of learning rate.
    lr_decay = 0.1
    # Epochs at which learning rate decays
    lr_decay_epoch = (20, 24)
    # Learning rate scheduler mode. options are step, poly and cosine
    lr_mode = 'step'
    # Number of iterations for warmup.
    lr_warmup = 500
    # Starging lr warmup factor.
    lr_warmup_factor = 1. / 3.
    # Gradient clipping.
    clip_gradient = -1
    # Momentum
    momentum = 0.9
    # Weight decay
    wd = 1e-4

    # RPN options
    # -----------
    # Filter top proposals before NMS in training of RPN.
    rpn_train_pre_nms = 12000
    # Return top proposal results after NMS in training of RPN.
    # Will be set to rpn_train_pre_nms if it is larger than rpn_train_pre_nms.
    rpn_train_post_nms = 2000
    # RPN box regression transition point from L1 to L2 loss.
    # Set to 0.0 to make the loss simply L1.
    rpn_smoothl1_rho = 0.001
    # Proposals whose size is smaller than ``min_size`` will be discarded.
    rpn_min_size = 1

    # R-CNN options
    # -------------
    # Number of samples for RPN targets.
    rcnn_num_samples = 512
    # Anchor with IOU larger than ``rcnn_pos_iou_thresh`` is regarded as positive samples.
    rcnn_pos_iou_thresh = 0.5
    # ``rcnn_pos_iou_thresh`` defines how many positive samples
    # (``rcnn_pos_iou_thresh * num_sample``) is to be sampled.
    rcnn_pos_ratio = 0.25
    # R-CNN box regression transition point from L1 to L2 loss.
    # Set to 0.0 to make the loss simply L1.
    rcnn_smoothl1_rho = 0.001

    # Misc
    # ----
    # log interval in terms of iterations
    log_interval = 100
    seed = 233
    # Whether to enable verbose logging
    verbose = False
    # Number of threads for executor for scheduling ops.
    # More threads may incur higher GPU memory footprint,
    # but may speed up throughput. Note that when horovod is used,
    # it is set to 1.
    executor_threads = 4


@dataclass
class ValidCfg:
    # Filter top proposals before NMS in testing of RPN.
    rpn_test_pre_nms = 6000
    # Return top proposal results after NMS in testing of RPN.
    # Will be set to rpn_test_pre_nms if it is larger than rpn_test_pre_nms.
    rpn_test_post_nms = 1000
    # Epoch interval for validation
    val_interval = 1

@dataclass
class MaskRCNNCfg:
    mask_rcnn : MaskRCNN = field(default_factory=MaskRCNN)
    train : TrainCfg = field(default_factory=TrainCfg)
    valid : ValidCfg = field(default_factory=ValidCfg)    # Dataset name. eg. 'coco', 'voc'
    dataset = 'coco'
    # Training with GPUs, you can specify (1,3) for example.
    gpus = (0,)
    # Resume from previously saved parameters if not None.
    # For example, you can resume from ./faster_rcnn_xxx_0123.params.
    resume = ''
    # Saving parameter prefix
    save_prefix = ''
    # Saving parameters epoch interval, best model will always be saved.
    save_interval = 1
    # Use MXNet Horovod for distributed training. Must be run with OpenMPI.
    horovod = False
    # Number of data workers, you can use larger number to accelerate data loading,
    # if your CPU and GPUs are powerful.
    num_workers = 16
    # KV store options. local, device, nccl, dist_sync, dist_device_sync,
    # dist_async are available.
    kv_store = 'nccl'
    # Whether to disable hybridize the model. Memory usage and speed will decrese.
    disable_hybridization = False
    # Use NVIDIA MSCOCO API. Make sure you install first.
    use_ext = False



from __future__ import print_function

import numpy as np
import gluoncv as gcv

def test_bbox_xywh_to_xyxy():
    # test list
    a = [20, 30, 100.2, 300.4]
    expected = [20, 30, 119.2, 329.4]
    np.testing.assert_allclose(gcv.utils.bbox.bbox_xywh_to_xyxy(a), expected)
    aa = np.array([a, a])
    bb = np.array([expected, expected])
    np.testing.assert_allclose(gcv.utils.bbox.bbox_xywh_to_xyxy(aa), bb)

if __name__ == '__main__':
    import nose
    nose.runmodule()



"""GluonCV neural network layers"""
# pylint: disable=wildcard-import
from __future__ import absolute_import

from . import bbox
from . import coder
from . import feature
from . import matcher
from . import predictor
from . import sampler
from .block import *
from .gn import *



import os
import json
import logging
import hashlib
import pickle
import numpy as np
from collections import defaultdict
from pathlib import Path

from .io.video_io import VideoSortedFolderReader, VideoFrameReader
from .utils.serialization_utils import save_json, load_json, ComplexEncoder, \
    save_pickle, load_pickle


logging.basicConfig(level=logging.INFO)
log = logging.getLogger()


class FieldNames:
    """
    Keys found in annotation dict
    """
    # Entity Fields
    TIME = 'time'
    LABELS = 'labels'
    ID = 'id'
    BLOB = 'blob'

    # Sample fields
    METADATA = 'metadata'
    ENTITY_LIST = 'entities'
    SAMPLE_FILE = 'sample_file'

    # Dataset fields
    SAMPLE_DICT = 'samples'
    SAMPLES = 'samples'
    DATASET_METADATA = 'metadata'
    DATASET_VERSION = 'version'
    CLASS_MAPPING = 'class_mapping'

    # Data fields
    DATE_ADDED = 'date_added'
    DURATION = 'duration'
    DATA_PATH = 'data_path'
    FILENAME = 'filename'
    BASE_DIR = 'base_dir'
    FILE_EXT = 'file_ext'
    FPS = 'fps'
    NUM_FRAMES = 'number_of_frames'
    RESOLUTION = 'resolution'
    WIDTH = 'width'
    HEIGHT = 'height'
    DATASET_SOURCE = 'data_source'
    SOURCE_ID = 'source_id'
    SAMPLE_SOURCE = 'sample_source'
    ORIG_ID = 'orig_id'
    TEMPORAL_SEGMENTS = 'temporal_segments'
    DESCRIPTION = 'description'
    BOUNDING_BOXES = 'bb'
    HEAD_BBOX = "bb_head"
    FACE_BBOX = "bb_face"
    MASK = 'mask'
    KEYPOINTS = 'keypoints'
    CONFIDENCE = 'confidence'
    DEFAULT_VALUES = 'default_values'
    DEFAULT_GT = 'default_gt'
    FORMAT_VERSION = 'format_version'
    DATE_MODIFIED = 'last_modified'
    CHANGES = 'changes'
    KEY_HASH = 'key_hash'
    FRAME_IDX = 'frame_idx'

    @classmethod
    def get_key_hash(cls):
        return 0


class SplitNames:
    TRAIN = "train"
    VAL = "val"
    TEST = "test"


def get_instance_list(cls, raw_entity_list):
    return [cls(raw_info=x) for x in raw_entity_list]


class AnnoEntity:
    """
    One annotation entity.
    A entity can refer to a video label, or a time segment annotation, or a frame annotation
    It has a required field "time", which should be in milliseconds
    Other than that, there are a few optional fields:
        id: if the entity has an id
        labels: if the entity has some categorical labels
        blob: other annotation information of this entity
    """

    def __init__(self, time=None, id=None, raw_info=None, validate=True):
        self._raw_info = raw_info
        self._time = None
        self._labels = None
        self._mask = None
        self._blob = {}

        self.id = None
        self.bbox = None
        self.keypoints = None
        self.confidence = None

        if raw_info is not None:
            if time is not None or id is not None:
                log.warning("time or id were specified, but raw_info was set and so they will be ignored")
            self._parse_info()
        else:
            if validate:
                self.time = time
            else:
                self._time = time
            self.id = id

    @property
    def time(self):
        return self._time

    @time.setter
    def time(self, time):
        if isinstance(time, (int, float)):
            assert time >= 0
            self._time = time
        elif isinstance(time, str):
            self._time = float(time)
        elif isinstance(time, tuple):
            assert len(time) == 2
            self._time = float(time[0]), float(time[1])

    @property
    def labels(self):
        return self._labels

    @labels.setter
    def labels(self, new_labels):
        if self._labels:
            # compare keys and report warning
            diff_keys = set(self._labels.keys()).difference(new_labels.keys())
            log.warning("updating the labels with different keys: {}".format(diff_keys))
        self._labels = new_labels


    @property
    def keypoints_xyv(self):
        return np.vstack([np.asarray(self.keypoints[0::3], dtype=np.float32),
                np.asarray(self.keypoints[1::3], dtype=np.float32),
                np.asarray(self.keypoints[2::3], dtype=np.float32)]).T

    @keypoints_xyv.setter
    def keypoints_xyv(self, kp_xyv):
        self.keypoints = list(np.asarray(kp_xyv).flatten())

    @property
    def mask(self):
        return self._mask

    @mask.setter
    def mask(self, mask):
        self._mask = mask

    @property
    def frame_num(self):
        if FieldNames.FRAME_IDX not in self.blob:
            return None
        return self.blob[FieldNames.FRAME_IDX]

    @frame_num.setter
    def frame_num(self, frame_num):
        self.blob[FieldNames.FRAME_IDX] = frame_num

    @property
    def blob(self):
        return self._blob

    @blob.setter
    def blob(self, new_blob):
        if self._blob:
            # compare keys and report warning
            diff_keys = set(self._blob.keys()).difference(new_blob.keys())
            log.warning("updating the blob with different keys: {}".format(diff_keys))
        self._blob = new_blob

    def to_dict(self):
        FN = FieldNames
        out = dict()

        out[FN.TIME] = self.time

        if self.id is not None:
            out[FN.ID] = self.id

        if self.labels is not None:
            out[FN.LABELS] = self.labels

        if self.confidence is not None:
            out[FN.CONFIDENCE] = self.confidence

        if self.bbox is not None:
            out[FN.BOUNDING_BOXES] = self.bbox

        if self.mask is not None:
            out[FN.MASK] = self.mask

        if self.keypoints is not None:
            out[FN.KEYPOINTS] = self.keypoints

        if self.blob:
            out[FN.BLOB] = self.blob

        return out

    def _parse_info(self):
        FN = FieldNames

        raw_info = self._raw_info

        self._time = raw_info[FN.TIME]

        self.id = raw_info.get(FN.ID, None)
        self.bbox = raw_info.get(FN.BOUNDING_BOXES, None)
        self.keypoints = raw_info.get(FN.KEYPOINTS, None)
        self.confidence = raw_info.get(FN.CONFIDENCE, 1.0)
        self._labels = raw_info.get(FN.LABELS, None)
        self._mask = raw_info.get(FN.MASK, None)
        self._blob = raw_info.get(FN.BLOB, {})

    ### helper functions ###
    @property
    def x(self):
        return self.bbox[0] if self.bbox else None
    @property
    def y(self):
        return self.bbox[1] if self.bbox else None
    @property
    def w(self):
        return self.bbox[2] if self.bbox else None
    @property
    def h(self):
        return self.bbox[3] if self.bbox else None


class DataReader:
    def __init__(self, data_sample, max_frame_deviation=1, fps=None):

        self._data_sample = data_sample
        self._data_path = data_sample.data_path
        self._frame_reader = None
        self._frame_iter = None

        if fps is None:
            fps = data_sample.fps
        self._fps = fps

        self._max_frame_deviation = max_frame_deviation
        self._period = 1.0 / fps * 1000
        self._time_diff_cutoff = int(self._max_frame_deviation * self._period) - 1

        self._frame_reader = self._data_sample.frame_reader

    def __iter__(self):
        self._frame_iter = iter(self._frame_reader)
        return self

    def __next__(self):
        frame, ts = next(self._frame_iter)
        if frame is None:
            raise StopIteration
        entities = self._data_sample.get_entities_near_time(int(ts), self._time_diff_cutoff)
        return frame, ts, entities

    def __getitem__(self, item):
        frame, ts = self._frame_reader[item]
        entities = self._data_sample.get_entities_near_time(int(ts), self._time_diff_cutoff)
        return frame, ts, entities


class DataSample:
    """
    One sample in the dataset. This can be a video file or an image file.
    It contains a list of entities as annotations.
    Each data sample can have some meta data. For example in videos, one can have
        FPS
        Duration
        Number of frames
        Source id
        ...
    """
    _NO_SERIALIZE_FIELDS = ("_dataset",)

    def __init__(self, id, raw_info=None, root_path=None, metadata=None, dataset=None):
        self._raw_info = raw_info
        self._id = id
        self._entities = []
        if metadata is not None:
            self._metadata = dict(metadata)
        else:
            self._metadata = {}
        self._dataset = dataset
        self._filepath = None
        self._lazy_loaded = False
        self._root_path = None
        self._data_root_path = None
        self._cache_root_path = None
        self._raw_entities = None
        self._time_entity_dict = None
        self._entity_times = None
        self._times_unsorted = True
        self._id_entity_dict = None
        self._frame_entity_dict = None
        self._label_entity_dict = None
        self._init_entity_fields()

        self.set_root_path(root_path)
        if self._raw_info:
            self._parse()

    def set_root_path(self, root_path):
        self._root_path = root_path
        if root_path:
            self._data_root_path = GluonCVMotionDataset.get_data_path_from_root(root_path)
            self._cache_root_path = GluonCVMotionDataset.get_cache_path_from_root(root_path)

    def _set_filepath(self, filepath, already_loaded=False):
        self._filepath = filepath
        self._lazy_loaded = already_loaded

    @property
    def id(self):
        return self._id

    @property
    def metadata(self):
        return self._metadata

    @metadata.setter
    def metadata(self, new_md):
        if self._metadata:
            # compare keys and report warning
            diff_keys = set(self._metadata.keys()).difference(new_md.keys())
            log.warning("updating the metadata with different keys: {}".format(diff_keys))
        self._metadata = new_md

    def _lazy_init(self):
        if self._filepath and not self._lazy_loaded:
            self._lazy_load()
        if (self._raw_entities is not None) and not self._entities:
            self._entities = get_instance_list(AnnoEntity, self._raw_entities)
            self._raw_entities = None
            self._init_entity_fields()

    def __len__(self):
        return self.metadata[FieldNames.NUM_FRAMES]

    @property
    def duration(self):
        return self.metadata[FieldNames.DURATION]

    @property
    def entities(self) -> [AnnoEntity]:
        self._lazy_init()
        return self._entities

    @property
    def data_relative_path(self):
        if FieldNames.DATA_PATH in self.metadata:
            data_path = self.metadata[FieldNames.DATA_PATH]
        if FieldNames.BASE_DIR in self.metadata:
            data_path = self.metadata[FieldNames.BASE_DIR]
        return data_path

    @data_relative_path.setter
    def data_relative_path(self, data_path):
        self.metadata[FieldNames.DATA_PATH] = data_path

    @property
    def data_path(self):
        data_path = self.data_relative_path
        if self._data_root_path:
            data_path = os.path.join(self._data_root_path, data_path)

        return data_path

    @property
    def frame_reader(self):
        data_path = self.data_path
        if os.path.isdir(data_path):
            frame_reader = VideoSortedFolderReader(data_path, self.fps)
        else:
            frame_reader = VideoFrameReader(data_path)
        return frame_reader

    def get_cache_file(self, cache_name, extension=''):
        rel_path = os.path.splitext(self.data_relative_path)[0] if os.path.isfile(self.data_path) else self.data_relative_path
        return os.path.join(self._cache_root_path, cache_name, rel_path + extension)

    @property
    def fps(self):
        fps = self.metadata.get(FieldNames.FPS, None)
        return fps

    @property
    def period(self):
        """Retrieves the period in milliseconds (1000 / fps), if fps is unset, returns None"""
        fps = self.fps
        return 1000 / fps if fps else None

    @property
    def width(self):
        width = self.metadata.get(FieldNames.RESOLUTION, {}).get(FieldNames.WIDTH, None)
        return width

    @property
    def height(self):
        height = self.metadata.get(FieldNames.RESOLUTION, {}).get(FieldNames.HEIGHT, None)
        return height

    @property
    def num_minutes(self):
        frames = self.metadata[FieldNames.NUM_FRAMES]
        fps = self.fps if self.fps > 1 else 30.
        return frames/fps/60.

    def get_data_reader(self):
        return DataReader(self)

    def _init_entity_fields(self):
        self._time_entity_dict = defaultdict(list)
        self._id_entity_dict = defaultdict(list)
        self._frame_entity_dict = defaultdict(list)
        self._label_entity_dict = defaultdict(list)
        self._entity_times = []
        for entity in self.entities:
            self._update_key_dicts(entity)

    @property
    def frame_num_entity_dict(self):
        self._lazy_init()
        return self._frame_entity_dict

    def get_entities_for_frame_num(self, frame_idx):
        return self.frame_num_entity_dict[frame_idx]

    @property
    def id_entity_dict(self):
        self._lazy_init()
        return self._id_entity_dict

    def get_entities_with_id(self, id):
        return self.id_entity_dict[id]

    @property
    def time_entity_dict(self):
        self._lazy_init()
        return self._time_entity_dict

    def get_entities_at_time(self, time):
        return self.time_entity_dict[time]

    @property
    def label_entity_dict(self):
        self._lazy_init()
        return self._label_entity_dict

    def get_entities_with_label(self, label):
        return self.label_entity_dict[label]

    def _get_entity_times_sorted(self):
        if self._times_unsorted:
            self._entity_times.sort()
            self._times_unsorted = False
        return self._entity_times

    def get_entities_near_time(self, time, time_diff_cutoff=None):
        import bisect
        self._lazy_init()

        if time_diff_cutoff is None:
            # Set to the period msec - 1
            time_diff_cutoff = int(1.0 / self.fps * 1000) - 1

        entity_times = self._get_entity_times_sorted()

        if not entity_times:
            return []

        insert_pos = bisect.bisect(entity_times, time)
        time_left = entity_times[max(insert_pos - 1, 0)]
        time_right = entity_times[min(insert_pos, len(entity_times)-1)]

        diff_left = abs(time_left - time)
        diff_right = abs(time_right - time)

        if diff_right < diff_left:
            closest_time = time_right
            diff = diff_right
        else:
            closest_time = time_left
            diff = diff_left

        if diff < time_diff_cutoff:
            entities = self._time_entity_dict[closest_time]
        else:
            entities = []

        return entities

    def _update_key_dicts(self, entity):
        self._id_entity_dict[entity.id].append(entity)
        new_time = entity.time not in self._time_entity_dict
        self._time_entity_dict[entity.time].append(entity)
        if entity.time is not None and new_time:
            self._entity_times.append(entity.time)
            self._times_unsorted = True
        # an entity could have multiple labels
        if entity.labels is not None:
            for k, v in entity.labels.items():
                self._label_entity_dict[k].append(entity)
        if entity.frame_num is not None:
            self._frame_entity_dict[entity.frame_num].append(entity)
        else:
            if entity.time is not None and self.fps:
                # Time in seconds * fps = frame_num
                frame_num = round((entity.time / 1000) * self.fps)
            else:
                frame_num = None
            self._frame_entity_dict[frame_num].append(entity)

    def add_entity(self, entities):
        """
        Add a new entity or a list of entities to the sample
        :param entities:
        :return:
        """
        self._lazy_init()

        if isinstance(entities, AnnoEntity):
            self._entities.append(entities)
            self._update_key_dicts(entities)
        else:
            self._entities.extend(entities)
            for entity in entities:
                self._update_key_dicts(entity)

    def get_copy_without_entities(self, new_id=None):
        """
        :return: A new DataSample with the same id and metadata but no entities
        """
        if new_id is None:
            new_id = self.id
        return DataSample(new_id, root_path=self._root_path, metadata=self.metadata, dataset=self._dataset)

    def filter_entities(self, filter_fn):
        """
        :param filter_fn: When true, keep the entity, otherwise omit it
        """
        new_sample = self.get_copy_without_entities()
        for entity in self.entities:
            if filter_fn(entity):
                new_sample.add_entity(entity)
        return new_sample

    def get_non_empty_frames(self, filter_fn=None, fps=0):
        """
          Return indexes of all valid frames with the specified fps,
          whose annotation exists
        """
        if fps == 0:
            fps = self.fps
        interval = int(np.ceil(self.fps / fps))

        frame_idxs = []
        for idx in range(0, len(self), interval):
            entities = self.get_entities_for_frame_num(idx)
            if filter_fn is not None:
                entities, _ = filter_fn(entities)
            if len(entities) > 0:
                frame_idxs.append(idx)
        return sorted(frame_idxs)

    def to_dict(self, include_id=False, lazy_load_format=False):
        """
        Dump the information in this sample to a dict
        :return:
        """
        out = dict()

        if lazy_load_format and self._filepath:
            if self._metadata:
                out[FieldNames.METADATA] = self._metadata
            out[FieldNames.SAMPLE_FILE] = self._filepath
            return out

        self._lazy_init()

        if include_id:
            out[FieldNames.ID] = self.id
        out[FieldNames.METADATA] = self.metadata
        out[FieldNames.ENTITY_LIST] = [x.to_dict() for x in self.entities]
        return out

    def dump(self, filename, indent=0, include_id=True, **kwargs):
        save_json(self.to_dict(include_id=include_id), filename, indent=indent, **kwargs)

    def _get_lazy_load_path(self):
        if not self._filepath or not self._dataset:
            raise ValueError("Cannot get lazy load path without a filepath and dataset")
        base_path = Path(self._dataset.anno_path).with_suffix("")
        if self._filepath is True:
            filepath = base_path / (self.id + ".json")
        else:
            filepath = base_path / self._filepath
        return filepath

    def _lazy_load(self):
        filepath = self._get_lazy_load_path()
        self._raw_info = load_json(filepath)
        if self.metadata and self._raw_info.get(FieldNames.METADATA) != self.metadata:
            log.info("metadata did not match lazy loaded value for sample: {}, ignoring loaded value, this should be"
                      " resolved next time you dump the dataset".format(self.id))
            self._raw_info[FieldNames.METADATA] = self.metadata
        self._parse()
        self._lazy_loaded = True

    def _dump_for_lazy_load(self):
        if not self._lazy_loaded:
            log.debug("nothing lazy loaded to dump, returning")
            return
        filepath = self._get_lazy_load_path()
        filepath.parent.mkdir(parents=True, exist_ok=True)
        self.dump(filepath, include_id=False)

    def clear_lazy_loaded(self, clear_metadata=False):
        if not self._lazy_loaded:
            log.debug("nothing lazy loaded so nothing to clear")
            return
        self._entities = []
        self._raw_info = None
        self._init_entity_fields()
        if clear_metadata:
            self._metadata = {}
        self._lazy_loaded = False

    @classmethod
    def load(cls, filename, **kwargs):
        data = load_json(filename)
        raw_info = data["raw_info"] if "raw_info" in data else data
        return cls(data[FieldNames.ID], raw_info=raw_info, **kwargs)

    def _parse(self):
        self._metadata = self._raw_info.get(FieldNames.METADATA, {})
        # Lazy load entities for speed when loading dataset
        self._raw_entities = self._raw_info.get(FieldNames.ENTITY_LIST, [])
        filepath = self._raw_info.get(FieldNames.SAMPLE_FILE)
        if filepath is not None:
            self._filepath = filepath

    def copy(self):
        new = pickle.loads(pickle.dumps(self))
        for f in self._NO_SERIALIZE_FIELDS:
            setattr(new, f, getattr(self, f))
        return new

    def __getstate__(self):
        # Used by pickle and deepcopy, this prevents trying to copy the whole dataset due to the dataset back reference
        return {k: v for k, v in vars(self).items() if k not in self._NO_SERIALIZE_FIELDS}

    def __setstate__(self, state):
        self.__dict__.update(state)
        for f in self._NO_SERIALIZE_FIELDS:
            setattr(self, f, None)

    def __enter__(self):
        self._lazy_init()
        return self

    def __exit__(self):
        self.clear_lazy_loaded()


class GluonCVMotionDataset:
    ANNO_DIR = "annotation"
    CACHE_DIR = "cache"
    DATA_DIR = "raw_data"

    _DEFAULT_ANNO_FILE = "anno.json"
    _DEFAULT_SPLIT_FILE = "splits.json"

    def __init__(self, annotation_file=None, root_path=None, split_file=None, load_anno=True):
        """
        GluonCVMotionDataset
        :param annotation_file: The path to the annotation file, either a full path or a path relative to the root
         annotation path (root_path/annotation/), defaults to 'anno.json'
        :param root_path: The root path of the dataset, containing the 'annotation', 'cache', and 'raw_data' folders.
         If left empty it will be inferred from the annotation_file path by searching up until the 'annotation' folder
         is found, then going one more level up
        :param split_file: The path to the split file relative to the annotation file. It will be relative to the root
         annotation path instead if it starts with './'
        :param load_anno: Whether to load the annotation file, will cause an exception if it is true and file does not
         exist. Set this to false if you are just trying to write a new annotation file for example in an ingestion
         script
        """

        # a dict of DataSample instances
        import indexed
        self._samples = indexed.IndexedOrderedDict()
        self._splits = {}
        self._metadata = {}

        if annotation_file is None:
            annotation_file = self._DEFAULT_ANNO_FILE
            log.info("Annotation file not provided, defaulting to '{}'".format(annotation_file))

        self._root_path = self._get_root_path(root_path, annotation_file)

        if self._root_path:
            if not os.path.isdir(self._root_path):
                raise ValueError("Expected root folder but was not found at: {}".format(self._root_path))

            self._anno_path = os.path.join(self._root_path, self.ANNO_DIR, annotation_file)
            self._data_path = self.get_data_path_from_root(self._root_path)
            self._cache_path = self.get_cache_path_from_root(self._root_path)

            if not os.path.isdir(self._data_path):
                raise ValueError("Expected data folder but was not found at: {}".format(self._data_path))
        else:
            log.warning('Root path was not set for dataset, this should only happen when loading a lone annotation'
                        ' file for inspection')
            self._anno_path = annotation_file
            self._data_path = None
            self._cache_path = None

        if load_anno:
            if os.path.exists(self._anno_path):
                log.info('Loading annotation file {}...'.format(self._anno_path))
                # load annotation file
                if self._get_pickle_path().exists():
                    log.info('Found pickle file, loading this instead')
                loaded_pickle = self._load_pickle()
                if not loaded_pickle:
                    self._parse_anno(self._anno_path)
                self._split_path = self._get_split_path(split_file, self._anno_path)
                self._load_split()
            else:
                raise ValueError(
                    "load_anno is true but the anno path does not exist at: {}".format(self._anno_path))
        else:
            log.info('Skipping loading for annotation file {}'.format(self._anno_path))
            self._split_path = self._get_split_path(split_file, self._anno_path)

    def __len__(self):
        return len(self._samples)

    def __contains__(self, item):
        return item in self._samples

    def __getitem__(self, item):
        return self._samples[item]

    def __iter__(self):
        for item in self._samples.items():
            yield item

    @classmethod
    def get_data_path_from_root(cls, root_path):
        return os.path.join(root_path, cls.DATA_DIR)

    @classmethod
    def get_cache_path_from_root(cls, root_path):
        return os.path.join(root_path, cls.CACHE_DIR)

    def _get_root_path(self, root_path, annotation_file):
        if root_path is None:
            dirpath = os.path.dirname(annotation_file)
            if self.ANNO_DIR in dirpath:
                while os.path.basename(dirpath) != self.ANNO_DIR and dirpath != '/':
                    dirpath = os.path.dirname(dirpath)
                root_path = os.path.abspath(os.path.dirname(dirpath))
                log.info("Dataset root path inferred to be: {}".format(root_path))
        return root_path

    def _get_split_path(self, split_file, anno_path):
        split_path = split_file
        if split_path is None:
            split_path = self._DEFAULT_SPLIT_FILE
        if not os.path.isabs(split_path):
            if split_path.startswith("./"):
                anno_dir = os.path.join(self._root_path, self.ANNO_DIR)
            else:
                anno_dir = os.path.dirname(anno_path)
            split_path = os.path.join(anno_dir, split_path)
            split_subpath = split_path.replace(self._root_path or "", "").lstrip(os.path.sep)
            log.info("Split subpath: {}".format(split_subpath))
        return split_path

    @property
    def iter_samples(self):
        """
        returns a iterator of samples
        :return:
        """
        return self._samples.items()

    @property
    def samples(self):
        return self._samples.items()

    @property
    def sample_ids(self):
        return self._samples.keys()

    @property
    def sample_values(self):
        return self._samples.values()

    def get_split_ids(self, splits=None):
        if splits is None:
            # Default is return all ids
            return self._samples.keys()
        if isinstance(splits, str):
            splits = [splits]

        all_ids = []
        for split in splits:
            if split not in self._splits:
                log.warning("Provided split: {} was not in dataset".format(split))
            else:
                all_ids.extend(self._splits[split])

        return all_ids

    def get_split_samples(self, splits=None):
        split_ids = self.get_split_ids(splits)
        samples = []
        for split_id in split_ids:
            if split_id in self._samples:
                samples.append((split_id, self._samples[split_id]))
            else:
                log.info(f"Dataset is missing sample: {split_id} in split {self.get_use_for_id(split_id)}, skipping")
        return samples

    @property
    def train_samples(self):
        return self.get_split_samples(SplitNames.TRAIN)

    @property
    def val_samples(self):
        return self.get_split_samples(SplitNames.VAL)

    @property
    def trainval_samples(self):
        return self.get_split_samples([SplitNames.TRAIN, SplitNames.VAL])

    @property
    def test_samples(self):
        return self.get_split_samples(SplitNames.TEST)

    @property
    def all_samples(self):
        samples = self.get_split_samples([SplitNames.TRAIN, SplitNames.VAL, SplitNames.TEST])
        if not len(samples):
            samples = self.samples
        return samples

    def get_use_for_id(self, id):
        for use in self._splits:
            if id in self._splits[use]:
                return use
        return None

    @property
    def version(self):
        return __version__

    @property
    def metadata(self):
        return self._metadata

    @property
    def name(self):
        if not self._root_path:
            return None
        return os.path.basename(self._root_path)

    @property
    def root_path(self):
        return self._root_path

    @property
    def anno_root_path(self):
        return os.path.join(self._root_path, self.ANNO_DIR)

    @property
    def cache_root_path(self):
        return self._cache_path

    @property
    def data_root_path(self):
        return self._data_path

    @property
    def anno_path(self):
        return self._anno_path

    def _get_anno_subpath(self, anno_path, with_ext):
        subpath = Path(anno_path).relative_to(self.anno_root_path)
        if not with_ext:
            subpath = subpath.with_suffix("")
        return str(subpath)

    def get_anno_subpath(self, with_ext=False):
        return self._get_anno_subpath(self._anno_path, with_ext)

    def get_anno_suffix(self):
        subpath = self.get_anno_subpath()
        return "_" + subpath.replace(os.sep, "_")

    @property
    def split_path(self):
        return self._split_path

    def get_split_subpath(self, with_ext=False):
        return self._get_anno_subpath(self._split_path, with_ext)

    def get_split_suffix(self):
        subpath = self.get_split_subpath()
        return "_" + subpath.replace(os.sep, "_")

    @metadata.setter
    def metadata(self, new_md):
        if self._metadata:
            # compare keys and report warning
            diff_keys = set(self._metadata.keys()).difference(new_md.keys())
            log.warning("updating the metadata with different keys: {}".format(diff_keys))
        self._metadata = new_md

    @property
    def description(self):
        return self._metadata.get(FieldNames.DESCRIPTION, "")

    @description.setter
    def description(self, description):
        self._metadata[FieldNames.DESCRIPTION] = description

    def add_sample(self, sample:DataSample, dump_directly=False):
        # create a new sample so it functions just as it would if it were loaded from disk
        new_sample = DataSample(sample.id, raw_info=sample.to_dict(), root_path=self.root_path, dataset=self)
        self._samples[sample.id] = new_sample
        if dump_directly:
            new_sample._set_filepath(True, already_loaded=True)
            new_sample._dump_for_lazy_load()
            new_sample.clear_lazy_loaded()
        return new_sample

    def dumps(self, encoder=ComplexEncoder, **kwargs):
        return json.dumps(self._to_dict(), cls=encoder, **kwargs)

    def dump(self, filename=None, indent=0, **kwargs):
        if filename is None:
            filename = self._anno_path
            anno_dir = os.path.dirname(self._anno_path)
            if not os.path.exists(anno_dir):
                try:
                    os.mkdir(anno_dir)
                except OSError:
                    pass
        save_json(self._to_dict(), filename, indent=indent, **kwargs)

    def _get_pickle_path(self):
        return Path(self._anno_path).with_suffix(".pkl")

    def _anno_mod_time(self):
        return Path(self._anno_path).stat().st_mtime

    def dump_pickle(self, filepath=None, **kwargs):
        if filepath is None:
            filepath = str(self._get_pickle_path())
        modified_time = self._anno_mod_time()
        to_pickle = {
            "_samples": self._samples,
            "_metadata": self._metadata,
            "_raw_info": self._raw_info,
            "modified_time": modified_time
        }
        save_pickle(to_pickle, filepath, **kwargs)

    def _load_pickle(self, filepath=None):
        import datetime
        if filepath is None:
            filepath = str(self._get_pickle_path())
        if not os.path.exists(filepath):
            return False
        log.info('Loading pickle file {}'.format(filepath))
        try:
            loaded_dict = load_pickle(filepath)
        except OSError:
            log.warning("Failed to load pickle")
            return False

        stored_time = loaded_dict["modified_time"]
        modified_time = self._anno_mod_time()
        if stored_time == modified_time:
            self._samples = loaded_dict["_samples"]
            self._metadata = loaded_dict["_metadata"]
            self._raw_info = loaded_dict["_raw_info"]
            return True
        else:
            log.info(('The pickle stored modification time did not match the annotation, so not loading,'
                     ' please remove and regenerate the pickle, renaming to .old').format(filepath))
            new_filepath = str(filepath) + ".old_" + str(datetime.datetime.now()).replace(" ", "_")
            try:
                Path(filepath).rename(new_filepath)
            except OSError:
                pass
            return False

    def _parse_anno(self, annotation_file):
        json_info = load_json(annotation_file)
        self._raw_info = json_info
        log.info("loaded anno json")

        # load metadata
        assert FieldNames.DATASET_METADATA in self._raw_info, \
            "key: {} should present in the annotation file, we only got {}".format(
                FieldNames.DATASET_METADATA, self._raw_info.keys())
        self._metadata.update(self._raw_info[FieldNames.DATASET_METADATA])

        # check key map hash
        key_hash = self._metadata.get(FieldNames.KEY_HASH, '')
        assert key_hash == FieldNames.get_key_hash(), "Key list not matching. " \
                                                      "Maybe this annoation file is created with other versions." \
                                                      "Current version is {}".format(self.version)

        # load samples
        sample_dict = self._raw_info.get(FieldNames.SAMPLE_DICT, dict())
        root_path = self.root_path
        for k in sorted(sample_dict.keys()):
            self._samples[k] = DataSample(k, sample_dict[k], root_path=root_path, dataset=self)
            # self._samples[k] = sample_dict[k]
        log.info("loaded {} samples".format(len(self._samples)))

    def _load_split(self):
        if not os.path.exists(self._split_path):
            log.warning("Split path {} not found, skipping loading".format(self._split_path))
            return
        self._splits = load_json(self._split_path)
        split_sample_nums = {k: len(v) for k, v in self._splits.items()}
        log.info("Loaded splits with # samples: {}".format(split_sample_nums))

    @property
    def splits(self):
        return dict(self._splits)

    @splits.setter
    def splits(self, split_dict):
        self._splits = split_dict

    def dump_splits(self, filename=None, indent=2):
        if filename is None:
            filename = self._split_path
        save_json(self._splits, filename, indent=indent)

    def _to_dict(self, dump_sample_files=True):
        # add the version information to metadata
        self._metadata[FieldNames.DATASET_VERSION] = self.version
        self._metadata[FieldNames.KEY_HASH] = FieldNames.get_key_hash()

        dump_dict = dict()
        dump_dict[FieldNames.DATASET_METADATA] = self._metadata
        all_samples_dict = {}
        for sample_id, sample in self._samples.items():
            sample_dict = sample.to_dict(lazy_load_format=True)
            if sample._lazy_loaded and dump_sample_files:
                sample._dump_for_lazy_load()
            all_samples_dict[sample_id] = sample_dict
        dump_dict[FieldNames.SAMPLE_DICT] = all_samples_dict
        return dump_dict


def get_resized_256_video_location(data_sample: DataSample) -> str:
    return get_resized_video_location(data_sample, 256)


def get_resized_video_location(data_sample: DataSample, short_edge_res:int) -> str:
    return data_sample.get_cache_file('rgb_{}_mp4'.format(short_edge_res), extension='.mp4')


def get_vis_video_location(data_sample: DataSample) -> str:
    return data_sample.get_cache_file('full_res_mp4', extension='.mp4')


def get_vis_thumb_location(data_sample: DataSample) -> str:
    return data_sample.get_cache_file('thumbnails', extension='.jpg')


def get_vis_gt_location(data_sample: DataSample, cache_subpath: str) -> str:
    return data_sample.get_cache_file(os.path.join('gt_vis_json', cache_subpath), extension='.json')


def get_gt_data_sample_location(data_sample: DataSample) -> str:
    return data_sample.get_cache_file('gt_data_sample_json', extension='.json')



# pylint: disable=abstract-method,unused-argument,arguments-differ,missing-docstring
"""Customized Layers.
"""
from __future__ import absolute_import
from mxnet import initializer
from mxnet.gluon import nn, contrib
from mxnet.gluon.nn import BatchNorm, HybridBlock

__all__ = ['BatchNormCudnnOff', 'Consensus', 'ReLU6', 'HardSigmoid', 'HardSwish', 'Identity']

class BatchNormCudnnOff(BatchNorm):
    """Batch normalization layer without CUDNN. It is a temporary solution.

    Parameters
    ----------
    kwargs : arguments goes to mxnet.gluon.nn.BatchNorm
    """
    def __init__(self, **kwargs):
        super(BatchNormCudnnOff, self).__init__(**kwargs)

    def hybrid_forward(self, F, x, gamma, beta, running_mean, running_var):
        return F.BatchNorm(x, gamma, beta, running_mean, running_var,
                           name='fwd', cudnn_off=True, **self._kwargs)

class Consensus(HybridBlock):
    """Consensus used in temporal segment networks.

    Parameters
    ----------
    nclass : number of classses
    num_segments : number of segments
    kwargs : arguments goes to mxnet.gluon.nn.Consensus
    """

    def __init__(self, nclass, num_segments, **kwargs):
        super(Consensus, self).__init__(**kwargs)
        self.nclass = nclass
        self.num_segments = num_segments

    def hybrid_forward(self, F, x):
        reshape_out = x.reshape((-1, self.num_segments, self.nclass))
        consensus_out = reshape_out.mean(axis=1)
        return consensus_out

class ReLU6(HybridBlock):
    """RelU6 used in MobileNetV2 and MobileNetV3.

    Parameters
    ----------
    kwargs : arguments goes to mxnet.gluon.nn.ReLU6
    """

    def __init__(self, **kwargs):
        super(ReLU6, self).__init__(**kwargs)

    def hybrid_forward(self, F, x):
        return F.clip(x, 0, 6, name="relu6")

class HardSigmoid(HybridBlock):
    """HardSigmoid used in MobileNetV3.

    Parameters
    ----------
    kwargs : arguments goes to mxnet.gluon.nn.HardSigmoid
    """
    def __init__(self, **kwargs):
        super(HardSigmoid, self).__init__(**kwargs)
        self.act = ReLU6()

    def hybrid_forward(self, F, x):
        return self.act(x + 3.) / 6.

class HardSwish(HybridBlock):
    """HardSwish used in MobileNetV3.

    Parameters
    ----------
    kwargs : arguments goes to mxnet.gluon.nn.HardSwish
    """
    def __init__(self, **kwargs):
        super(HardSwish, self).__init__(**kwargs)
        self.act = HardSigmoid()

    def hybrid_forward(self, F, x):
        return x * self.act(x)

class SoftmaxHD(HybridBlock):
    """Softmax on multiple dimensions

    Parameters
    ----------
    axis : the axis for softmax normalization
    """
    def __init__(self, axis=(2, 3), **kwargs):
        super(SoftmaxHD, self).__init__(**kwargs)
        self.axis = axis

    def hybrid_forward(self, F, x):
        x_max = F.max(x, axis=self.axis, keepdims=True)
        x_exp = F.exp(F.broadcast_minus(x, x_max))
        norm = F.sum(x_exp, axis=self.axis, keepdims=True)
        res = F.broadcast_div(x_exp, norm)
        return res

class DSNT(HybridBlock):
    '''DSNT module to translate heatmap to coordinates

    Parameters
    ----------
    size : int or tuple,
        (width, height) of the input heatmap
    norm : str, the normalization method for heatmap
        available methods are 'softmax', or 'sum'
    axis : the axis for input heatmap
    '''
    def __init__(self, size, norm='sum', axis=(2, 3), **kwargs):
        super(DSNT, self).__init__(**kwargs)
        if isinstance(size, int):
            self.size = (size, size)
        else:
            self.size = size
        self.axis = axis
        self.norm = norm
        if self.norm == 'softmax':
            self.softmax = SoftmaxHD(self.axis)
        elif self.norm != 'sum':
            raise ValueError("argument `norm` only accepts 'softmax' or 'sum'.")

        self.wfirst = 1 / (2 * self.size[0])
        self.wlast = 1 - 1 / (2 * self.size[0])
        self.hfirst = 1 / (2 * self.size[1])
        self.hlast = 1 - 1 / (2 * self.size[1])

    def hybrid_forward(self, F, M):
        # pylint: disable=missing-function-docstring
        if self.norm == 'softmax':
            Z = self.softmax(M)
        elif self.norm == 'sum':
            norm = F.sum(M, axis=self.axis, keepdims=True)
            Z = F.broadcast_div(M, norm)
        else:
            Z = M
        x = F.linspace(self.wfirst, self.wlast, self.size[0]).expand_dims(0)
        y = F.linspace(self.hfirst, self.hlast, self.size[1]).expand_dims(0).transpose()
        output_x = F.sum(F.broadcast_mul(Z, x), axis=self.axis)
        output_y = F.sum(F.broadcast_mul(Z, y), axis=self.axis)
        res = F.stack(output_x, output_y, axis=2)
        return res, Z

class DUC(HybridBlock):
    '''Upsampling layer with pixel shuffle
    '''
    def __init__(self, planes, upscale_factor=2, **kwargs):
        super(DUC, self).__init__(**kwargs)
        self.conv = nn.Conv2D(planes, kernel_size=3, padding=1, use_bias=False)
        self.bn = BatchNormCudnnOff(gamma_initializer=initializer.One(),
                                    beta_initializer=initializer.Zero())
        self.relu = nn.Activation('relu')
        self.pixel_shuffle = contrib.nn.PixelShuffle2D(upscale_factor)

    def hybrid_forward(self, F, x):
        x = self.conv(x)
        x = self.bn(x)
        x = self.relu(x)
        x = self.pixel_shuffle(x)
        return x

class Identity(HybridBlock):
    """Block that passes through the input directly.
    This block can be used in conjunction with HybridConcatenate
    block for residual connection.
    Example::
        net = HybridConcatenate()
        net.add(nn.Dense(10, activation='relu'))
        net.add(nn.Dense(20))
        net.add(Identity())
    """
    def __init__(self):
        super(Identity, self).__init__()

    def hybrid_forward(self, F, x):
        return x



import os
import glob
import subprocess
import tempfile
from pathlib import Path
import cv2
import numpy as np
from PIL import Image


class ColorSpace(object):
    RGB = 0
    BGR = 3
    GRAY = 2


convert_from_to_dict = {ColorSpace.BGR: {ColorSpace.RGB: cv2.COLOR_BGR2RGB,
                                         ColorSpace.GRAY: cv2.COLOR_BGR2GRAY},
                        ColorSpace.RGB: {ColorSpace.BGR: cv2.COLOR_RGB2BGR,
                                         ColorSpace.GRAY: cv2.COLOR_RGB2GRAY},
                        ColorSpace.GRAY: {ColorSpace.BGR: cv2.COLOR_GRAY2BGR,
                                          ColorSpace.RGB: cv2.COLOR_GRAY2RGB}}

FFMPEG_FOURCC = {
    'libx264': 0x21,
    'avc1': cv2.VideoWriter_fourcc(*'avc1'),
    'mjpeg': 0x6c,
    'mpeg-4': 0x20
}


def convert_color_from_to(frame, cs_from, cs_to):
    if cs_from not in convert_from_to_dict or cs_to not in convert_from_to_dict[cs_from]:
        raise Exception('color conversion is not supported')
    convert_spec = convert_from_to_dict[cs_from][cs_to]
    return cv2.cvtColor(frame, convert_spec)


def read_vid_rgb(file):
    cap = cv2.VideoCapture(file)
    all_ts = []
    all_frames = []
    while True:
        ts = int(cap.get(cv2.CAP_PROP_POS_MSEC))
        frame = read_frame(cap)
        if frame is None:
            break
        all_frames.append(frame)
        all_ts.append(ts)

    fps = cap.get(cv2.CAP_PROP_FPS)
    return InMemoryVideo(all_frames, fps, frame_ts=all_ts)


def format_frame(frame, color_space=ColorSpace.RGB):
    if color_space != ColorSpace.BGR:
        frame = convert_color_from_to(frame, ColorSpace.BGR, color_space)
    return frame


def read_frame(cap):
    _, frame = cap.read()
    if frame is None:
        return frame
    return Image.fromarray(format_frame(frame, ColorSpace.RGB), 'RGB')


def read_img(file):
    frame = cv2.imread(file)
    if frame is None:
        return frame
    return Image.fromarray(format_frame(frame, ColorSpace.RGB), 'RGB')


def write_img(file, img, color_space=ColorSpace.RGB):
    img = convert_color_from_to(img, color_space, ColorSpace.BGR)
    cv2.imwrite(file, img)


class VideoBaseClass(object):
    def __init__(self):
        raise NotImplementedError()

    def __del__(self):
        raise NotImplementedError()

    def __len__(self):
        raise NotImplementedError()

    def _set_frame_ndx(self, frame_num):
        raise NotImplementedError()

    def get_next_frame_time_stamp(self):
        raise NotImplementedError()

    def read(self):
        raise NotImplementedError()

    def __iter__(self):
        self._set_frame_ndx(0)
        return self

    def iter_frame_ts(self, start_ts=0):
        return FrameTimeStampIterator(self, start_ts)

    def next(self):
        return self.__next__()

    def __next__(self):
        ts = self.get_next_frame_time_stamp()
        frame = self.read()
        if frame is None:
            raise StopIteration()
        return frame, ts

    def __getitem__(self, frame_num):
        if self._next_frame_to_read != frame_num:
            self._set_frame_ndx(frame_num)
        ts = self.get_next_frame_time_stamp()
        return self.read(), ts

    @property
    def verified_len(self):
        return len(self)

    @property
    def fps(self):
        return self.get_frame_rate()

    @property
    def width(self):
        return self.get_width()

    @property
    def height(self):
        return self.get_height()

    def get_frame_ind_for_time(self, time_stamp):
        """
        Returns the index for the frame at the timestamp provided.
        The frame index returned is the first frame that occurs before or at the timestamp given.

        Args:
            time_stamp (int): the millisecond time stamp for the desired frame

        Returns (int):
            the index for the frame at the given timestamp.

        """
        assert isinstance(time_stamp, int)
        return int(self.fps * time_stamp / 1000.)

    def get_frame_for_time(self, time_stamp):
        return self[self.get_frame_ind_for_time(time_stamp)]

    def get_frame_rate(self):
        raise NotImplementedError()

    def get_width(self):
        raise NotImplementedError()

    def get_height(self):
        raise NotImplementedError()

    @property
    def duration(self):
        raise NotImplementedError()

    def asnumpy_and_ts(self):
        out = []
        out_ts = []
        for frame, ts in self.iter_frame_ts():
            out.append(frame)
            out_ts.append(ts)
        return out, out_ts

    def asnumpy(self):
        out = []
        for frame in self:
            out.append(frame)
        return out

    def num_frames(self):
        return len(self)

    def get_frame(self, index):
        return self[index]

    def get_frame_batch(self, index_list):
        '''
        Return a list of PIL Image classes
        Args:
            index_list (List[int]): list of indexes
            color_mode (str):  color mode of the pil image typically 'RGB'

        Returns: List[PIL.Image]

        '''
        return [self.get_frame(i) for i in index_list]


class FrameTimeStampIterator(object):
    def __init__(self, frame_reader, start_ts=0):
        self.frame_reader = frame_reader
        self.frame_reader._set_frame_time(start_ts)

    def __iter__(self):
        return self

    def next(self):
        return self.__next__()

    def __next__(self):
        return next(self.frame_reader)


class InMemoryVideo(VideoBaseClass):
    def __init__(self, frames=None, fps=None, frame_ts=None):
        self._frames = []
        if frames is not None:
            self._frames = list(frames)

        self._fps = fps
        self._next_frame_to_read = 0

        self._frame_ts = []
        if len(self._frames) > 0:
            assert len(frame_ts) == len(self._frames)
            assert all(a <= b for a, b in zip(frame_ts[:-1], frame_ts[1:]))
            self._frame_ts = frame_ts

    def __del__(self):
        pass

    def __len__(self):
        return len(self._frames)

    def _set_frame_ndx(self, frame_num):
        self._next_frame_to_read = frame_num

    def get_next_frame_time_stamp(self):
        if self._next_frame_to_read >= len(self._frame_ts):
            return None
        return self._frame_ts[self._next_frame_to_read]

    def read(self):
        if self._next_frame_to_read >= len(self._frames):
            return None
        f = self._frames[self._next_frame_to_read]
        self._next_frame_to_read += 1
        return f

    def __setitem__(self, key, value):
        self._next_frame_to_read = key + 1
        self._frames[key] = value

    def append(self, frame, ts=None):
        assert ts is None or len(self._frame_ts) == 0 or ts > self._frame_ts[-1]
        self._frames.append(frame)
        self._next_frame_to_read = len(self._frames)
        if ts is None:
            if len(self._frame_ts) > 0:
                self._frame_ts.append(self._frame_ts[-1] + 1000. / self.fps)
            else:
                self._frame_ts.append(0.)
        else:
            self._frame_ts.append(ts)

    def extend(self, frames, tss):
        assert all(a <= b for a, b in zip(tss[:-1], tss[1:]))
        self._frames.extend(frames)
        self._frame_ts.extend(tss)
        self._next_frame_to_read = len(self._frames)

    def get_frame_rate(self):
        return self._fps

    def asnumpy(self):
        return self._frames

    def get_frame_ind_for_time(self, time_stamp):
        ind = np.searchsorted(self._frame_ts, time_stamp)
        if ind > 0:
            ind -= 1
        return ind


class InMemoryMXVideo(InMemoryVideo):
    def asnumpy(self):
        return [f.asnumpy() for f in self._frames]


img_exts = ['.jpg', '.jpeg', '.jp', '.png']
vid_exts = ['.avi', '.mpeg', '.mp4', '.mov']


class VideoFrameReader(VideoBaseClass):
    def __init__(self, file):
        self.cap = None
        self.file_name = file
        self._next_frame_to_read = 0
        self._verified_len = None
        self.frame_cache = {}
        self._is_vid = None
        self._is_img = None
        self._len = None
        self._duration = None

    def __del__(self):
        if self.cap is not None:
            self.cap.release()

    @property
    def is_video(self):
        return not self.is_img

    @property
    def is_img(self):
        if self._is_img is None:
            _, ext = os.path.splitext(self.file_name)
            self._is_img = ext.lower() in img_exts
        return self._is_img

    def _lazy_init(self):
        if self.is_video and self.cap is None:
            self.cap = cv2.VideoCapture(self.file_name)

    def read_from_mem_cache(self):
        return None

    def read(self):
        self._lazy_init()
        if (not self.is_img) and self._next_frame_to_read != self.cap.get(cv2.CAP_PROP_POS_FRAMES):
            raise Exception("failed read frame check, stored {} , cap val {} , file {}".format(
                self._next_frame_to_read, self.cap.get(cv2.CAP_PROP_POS_FRAMES), self.file_name))
        if self.is_video:
            frame = read_frame(self.cap)
        else:
            if self._next_frame_to_read == 0:
                frame = read_img(self.file_name)
            else:
                frame = None
        if frame is None:
            self._verified_len = self._next_frame_to_read
        self._next_frame_to_read += 1
        return frame

    def _set_frame_ndx(self, frame_num):
        self._lazy_init()
        if self.is_video:
            self.cap.set(cv2.CAP_PROP_POS_FRAMES, frame_num)
        self._next_frame_to_read = frame_num

    def _set_frame_time(self, frame_ts):
        self._lazy_init()
        self.cap.set(cv2.CAP_PROP_POS_MSEC, frame_ts)
        self._next_frame_to_read = self.cap.get(cv2.CAP_PROP_POS_FRAMES)

    def get_frame_for_time(self, time_stamp):
        self._lazy_init()
        if self.is_video:
            self.cap.set(cv2.CAP_PROP_POS_MSEC, time_stamp)
            self._next_frame_to_read = self.cap.get(cv2.CAP_PROP_POS_FRAMES)
        return self.read()

    def get_next_frame_time_stamp(self):
        self._lazy_init()
        if self.is_video:
            return max(0, int(self.cap.get(cv2.CAP_PROP_POS_MSEC)))
        else:
            return 0

    def _init_len_and_duration(self):
        if self._duration is None:
            self._lazy_init()
            pos = self.cap.get(cv2.CAP_PROP_POS_MSEC)
            self.cap.set(cv2.CAP_PROP_POS_AVI_RATIO, 1)
            self._duration = int(self.cap.get(cv2.CAP_PROP_POS_MSEC))
            self._len = int(self.cap.get(cv2.CAP_PROP_POS_FRAMES))
            self.cap.set(cv2.CAP_PROP_POS_MSEC, pos)

    def __len__(self):
        if self.is_video:
            self._init_len_and_duration()
            return self._len
        else:
            return 1

    @property
    def duration(self):
        self._init_len_and_duration()
        return self._duration

    @property
    def verified_len(self):
        if self.is_video:
            return self._verified_len
        else:
            return 1

    def get_frame_rate(self):
        self._lazy_init()
        if self.is_video:
            return self.cap.get(cv2.CAP_PROP_FPS)
        else:
            return 1

    def get_width(self):
        self._lazy_init()
        return self.cap.get(cv2.CAP_PROP_FRAME_WIDTH)

    def get_height(self):
        self._lazy_init()
        return self.cap.get(cv2.CAP_PROP_FRAME_HEIGHT)


class VideoSortedFolderReader(VideoBaseClass):
    def __init__(self, data_path, fps, glob_pattern="*"):
        self._data_path = data_path

        self._glob_pattern = glob_pattern
        frame_paths = glob.glob(os.path.join(data_path, glob_pattern))
        self._frame_paths = sorted(frame_paths)

        self._next_frame_to_read = 0
        self._last_read_frame = None
        self._fps = fps
        self._period = 1.0 / fps * 1000

    def __del__(self):
        pass

    def __len__(self):
        return len(self._frame_paths)

    @property
    def duration(self):
        return round(self._period * len(self))

    def get_frame_rate(self):
        return self._fps

    def _set_frame_ndx(self, frame_num):
        self._next_frame_to_read = frame_num

    def _set_frame_time(self, frame_ts):
        self._set_frame_ndx(round(frame_ts / self._period))

    def get_next_frame_time_stamp(self):
        return int(self._next_frame_to_read * self._period)

    def read(self):
        read_idx = self._next_frame_to_read
        if read_idx >= len(self._frame_paths):
            return None
        frame = read_img(self._frame_paths[read_idx])
        self._last_read_frame = read_idx
        self._next_frame_to_read += 1
        return frame

    def get_image_ext(self):
        return Path(self._frame_paths[0]).suffix

    def get_frame_path(self, frame_num=None):
        if frame_num is None:
            frame_num = self._last_read_frame
        return self._frame_paths[frame_num]


def write_video_rgb(file, frames, fps=None):
    # check if data has the fps property (eg: InMemoryVideo, VideoFrameReader or VideoCacheReader)
    if fps is None:
        fps = 30
    try:
        fps = frames.fps
    except:
        pass

    # write the video data frame-by-frame
    writer = None
    for frame in frames:
        frame = np.asarray(frame)
        if writer is None:
            writer = cv2.VideoWriter(file, FFMPEG_FOURCC['libx264'],
                                     fps=fps, frameSize=frame.shape[1::-1],
                                     isColor=True)
        frame = cv2.cvtColor(frame, cv2.COLOR_RGB2BGR)
        writer.write(frame)

    if writer is not None:
        writer.release()


def cmd_with_addn_args(cmd, addn_args):
    if addn_args:
        if isinstance(addn_args, str):
            addn_args = addn_args.split()
        return cmd + addn_args
    else:
        return cmd


def resize_and_write_video_ffmpeg(in_path, out_path, short_edge_res,
                                  scaling_algorithm="lanczos", raw_scale_input=None,
                                  keep_audio=True, addn_args=None):
    # See https://trac.ffmpeg.org/wiki/Scaling for scaling options / details

    if short_edge_res is not None and raw_scale_input is not None:
        raise ValueError("Either short_edge_res or raw_scale_input should be provided, not both")

    if short_edge_res is not None:
        # The input height, divided by the minimum of the width and height 
        # (so either = 1 or > 1) times the new short edge, 
        # then round to the nearest 2. 
        # We keep the aspect ratio of the width and make sure it is also divisible
        # by 2 by using '-2' (see the ffpeg scaling wiki)
        scale_arg = "-2:'round( ih/min(iw,ih) * {} /2)*2'".format(short_edge_res)

        # Alternatively:
        # scale_arg = "{res}:{res}:force_original_aspect_ratio=increase".format(res=short_edge_res)
        # In case the output has a non even dimension (e.g. 301) after rescaling, 
        # we crop the single extra pixel
        # crop_arg = "floor(iw/2)*2:floor(ih/2)*2"
    else:
        scale_arg = raw_scale_input

    if keep_audio:
        audio_arg = None
    else:
        audio_arg = "-an"

    scale_arg += ":flags={}".format(scaling_algorithm)

    with tempfile.TemporaryDirectory() as tmp_path:
        tmp_file_path = os.path.join(tmp_path, os.path.basename(out_path))

        ffmpeg_cmd = ["ffmpeg", "-y", "-hide_banner", "-loglevel", "error",
                      "-i", in_path, "-vf", "scale={}".format(scale_arg),
                      audio_arg]
        ffmpeg_cmd = cmd_with_addn_args(ffmpeg_cmd, addn_args)
        ffmpeg_cmd += ["-strict", "experimental", tmp_file_path]
        ffmpeg_cmd = [x for x in ffmpeg_cmd if x is not None]

        subprocess.run(ffmpeg_cmd, check=True)

        subprocess.run(["mv", tmp_file_path, out_path], check=True)


def resize_and_write_video(file, frames, short_edge_res, fps=None):
    # check if data has the fps property (eg: InMemoryVideo, VideoFrameReader or VideoCacheReader)
    if fps is None:
        fps = 30
    try:
        fps = frames.fps
    except:
        pass

    # write the video data frame-by-frame
    writer = None
    new_size = None
    for frame in frames:
        if new_size is None:
            factor = float(short_edge_res) / min(frame.size)
            new_size = [int(i * factor) for i in frame.size]

        frame_np = frame.resize(new_size)
        frame_np = np.asarray(frame_np)

        if writer is None:
            writer = cv2.VideoWriter(file, FFMPEG_FOURCC['libx264'],
                                     fps=fps, frameSize=frame_np.shape[1::-1],
                                     isColor=True)
        frame_np = cv2.cvtColor(frame_np, cv2.COLOR_RGB2BGR)
        writer.write(frame_np)

    if writer is not None:
        writer.release()


def write_img_files_to_vid(out_file, in_files, fps=None):
    # check if data has the fps property (eg: InMemoryVideo, VideoFrameReader or VideoCacheReader)
    if fps is None:
        fps = 30

    # write the video data frame-by-frame
    writer = None
    for in_file in in_files:
        with open(in_file, 'rb') as fp:
            frame = Image.open(fp)
            frame = np.asarray(frame)
            frame = cv2.cvtColor(frame, cv2.COLOR_RGB2BGR)
            if writer is None:
                writer = cv2.VideoWriter(out_file, FFMPEG_FOURCC['avc1'],
                                         fps=fps, frameSize=frame.shape[1::-1],
                                         isColor=True)
            writer.write(frame)

    if writer is not None:
        writer.release()


def write_img_files_to_vid_ffmpeg(out_file, in_files, fps=None):
    if fps is None:
        fps = 30
    input_str = "'\nfile '".join(in_files)
    input_str = "file '" + input_str + "'\n"

    with tempfile.TemporaryDirectory() as tmp_path:
        tmp_file_path = os.path.join(tmp_path, os.path.basename(out_file))
        # See https://trac.ffmpeg.org/wiki/Slideshow 
        # for why we are using input_str like this (for concat filter)
        # Need -safe 0 due to:
        # https://stackoverflow.com/questions/38996925/ffmpeg-concat-unsafe-file-name
        ret = subprocess.run(["ffmpeg", "-y", "-hide_banner", "-loglevel", "error",
                              "-f", "concat", "-safe", "0", "-r", str(fps), "-i", "/dev/stdin",
                              tmp_file_path],
                             input=input_str.encode('utf-8'), check=True)
        subprocess.run(["mv", tmp_file_path, out_file], check=True)
    return ret


def convert_vid_ffmpeg(in_path, out_path, addn_args=None):
    # muxing queue size bug workaround:
    cmd = ["ffmpeg", "-y", "-hide_banner", "-loglevel", "error",
           "-i", in_path, "-max_muxing_queue_size", "99999"]
    cmd = cmd_with_addn_args(cmd, addn_args)
    with tempfile.TemporaryDirectory() as tmp_path:
        tmp_file_path = os.path.join(tmp_path, os.path.basename(out_path))
        cmd += [tmp_file_path]
        ret = subprocess.run(cmd, check=True)
        subprocess.run(["mv", tmp_file_path, out_path], check=True)
    return ret



"""
Network utiliy functions
"""
# pylint: disable=line-too-long,unused-argument,missing-function-docstring
from __future__ import absolute_import
import logging

import mxnet as mx
from mxnet.gluon import HybridBlock
from mxnet.gluon import nn
from mxnet.gluon.contrib import nn as contrib_nn

from .mobilenet import get_mobilenet, Mish


class FPNFeatureExpander(HybridBlock):
    """Feature extractor with additional layers to append.
    This is specified for ``Feature Pyramid Network for Object Detection``
    which implement ``Top-down pathway and lateral connections``.

    Parameters
    ----------
    network : str or HybridBlock or Symbol
        Logic chain: load from gluon.model_zoo.vision if network is string.
        Convert to Symbol if network is HybridBlock.
    outputs : str or list of str
        The name of layers to be extracted as features
    num_filters : list of int e.g. [256, 256, 256, 256]
        Number of filters to be appended.
    use_1x1 : bool
        Whether to use 1x1 convolution
    use_upsample : bool
        Whether to use upsample
    use_elewadd : float
        Whether to use element-wise add operation
    use_p6 : bool
        Whther use P6 stage, this is used for RPN experiments in ori paper
    no_bias : bool
        Whether use bias for Convolution operation.
    norm_layer : HybridBlock or SymbolBlock
        Type of normalization layer.
    norm_kwargs : dict
        Arguments for normalization layer.
    pretrained : bool
        Use pretrained parameters as in gluon.model_zoo if `True`.
    ctx : Context
        The context, e.g. mxnet.cpu(), mxnet.gpu(0).
    inputs : list of str
        Name of input variables to the network.

    """

    def __init__(self, network, outputs, ssd_filters, fpn_filters, use_1x1=True, use_upsample=True,
                 use_elewadd=True, use_p6=False, no_bias=True, pretrained=False, norm_layer='BatchNorm',
                 norm_kwargs=None, ctx=mx.cpu(), inputs=('data',), use_bn=True, reduce_ratio=1.0,
                 min_depth=128, global_pool=False, use_mish=False, **kwargs):
        super(FPNFeatureExpander, self).__init__(**kwargs)

        gluon_norm_kwargs = {}
        if norm_kwargs is not None and 'num_devices' in norm_kwargs:
            gluon_norm_kwargs = {'num_devices': norm_kwargs['num_devices']}

        if norm_layer == nn.BatchNorm:
            logging.info("use BatchNorm for the backbone model.")
            network = get_mobilenet(1.0, pretrained=pretrained, ctx=ctx, mish=use_mish)
        else:
            logging.info("use SyncBatchNorm for the backbone model.")
            network = get_mobilenet(1.0, pretrained=pretrained, ctx=ctx,
                                    norm_layer=mx.gluon.contrib.nn.SyncBatchNorm,
                                    norm_kwargs=gluon_norm_kwargs, mish=use_mish)

        self.output_num = len(fpn_filters)
        self.norm_layer = norm_layer
        net = network.features
        with self.name_scope():
            self.ssd_backbone = nn.HybridSequential()
            self.build_output(net, [0, 21])
            self.build_output(net, [21, 33])
            self.build_output(net, [33, 69])
            self.build_output(net, [69, 81])
            for num_filter in ssd_filters:
                self.ssd_backbone.add(build_down_sample(num_filter, 2, self.norm_layer, gluon_norm_kwargs, use_mish=use_mish))

            self.upper_feat = nn.HybridSequential()
            self.lower_feat = nn.HybridSequential()
            self.fuse_feat = nn.HybridSequential()
            self.deconv_feat = nn.HybridSequential()

            self.fuse_feat_late = nn.HybridSequential()
            self.fpn_conv = nn.HybridSequential()
            self.downsample_conv = nn.HybridSequential()

            for idx in range(self.output_num):
                self.upper_feat.add(convolution(fpn_filters[idx], 1, self.norm_layer, gluon_norm_kwargs, pad=0, use_mish=use_mish))
                self.deconv_feat.add(deform_convolution(fpn_filters[idx], self.norm_layer, gluon_norm_kwargs, use_mish=use_mish))
                self.fuse_feat_late.add(convolution(fpn_filters[idx], 3, self.norm_layer, gluon_norm_kwargs, use_mish=use_mish))
                if idx > 0:
                    self.lower_feat.add(convolution(fpn_filters[idx], 1, self.norm_layer, gluon_norm_kwargs, pad=0, use_mish=use_mish))
                    self.fuse_feat.add(convolution(fpn_filters[idx], 3, self.norm_layer, gluon_norm_kwargs, use_mish=use_mish))
                    self.fpn_conv.add(convolution(fpn_filters[idx], 3, self.norm_layer, gluon_norm_kwargs, use_mish=use_mish))
                    self.downsample_conv.add(convolution(fpn_filters[idx], 3, self.norm_layer,
                                                         gluon_norm_kwargs, stride=2, pad=1, use_mish=use_mish))

    def hybrid_forward(self, F, x):
        forward_feature = []
        for i in range(self.output_num):
            x = self.ssd_backbone[i](x)
            forward_feature.append(x)

        aggregate_feat = []
        trans_feat = []
        for i in range(len(forward_feature) - 1, -1, -1):
            if i == len(forward_feature) - 1:
                P_fuse = self.upper_feat[i](forward_feature[i])
                trans_feat.append(P_fuse)
                aggregate_feat.append(P_fuse)
                continue
            P_up = self.upper_feat[i](
                F.UpSampling(aggregate_feat[-1], scale=2, sample_type='nearest', workspace=512, num_args=1))
            P_low = self.lower_feat[i](forward_feature[i])
            trans_feat.append(P_low)
            P_clip = F.Crop(*[P_up, P_low])
            P_sum = F.ElementWiseSum(*[P_clip, P_low])
            P_fuse = self.fuse_feat[i](P_sum)
            aggregate_feat.append(P_fuse)

        bi_fpn_feat = []
        for i in range(len(aggregate_feat) - 1, -1, -1):
            if i == len(aggregate_feat) - 1:
                P_fuse_late = self.fuse_feat_late[i](aggregate_feat[i])
                bi_fpn_feat.append(P_fuse_late)
                continue
            f_down = self.downsample_conv[i](bi_fpn_feat[-1])
            f_up = self.fpn_conv[i](aggregate_feat[i])
            f_clip = F.Crop(*[f_down, f_up])
            f_sum_0 = F.ElementWiseSum(*[f_clip, f_up])
            if i == 0:
                f_fuse = self.fuse_feat_late[i](f_sum_0)
                bi_fpn_feat.append(f_fuse)
                continue
            f_clip_sum = F.Crop(*[f_sum_0, trans_feat[i]])
            f_sum_1 = F.ElementWiseSum(*[f_clip_sum, trans_feat[i]])
            f_fuse = self.fuse_feat_late[i](f_sum_1)
            bi_fpn_feat.append(f_fuse)

        deform_feat = []
        for i, _ in enumerate(bi_fpn_feat):
            deform_feat.append(self.deconv_feat[i](bi_fpn_feat[i]))

        return deform_feat, bi_fpn_feat

    def build_output(self, net, layer_range):
        assert len(layer_range) == 2
        branch = nn.HybridSequential()
        for idx in range(layer_range[0], layer_range[1]):
            branch.add(net[idx])
        self.ssd_backbone.add(branch)


def build_down_sample(num_filters, stride, norm_layer, norm_kwargs, use_mish=False):
    """stack two Conv-BatchNorm-Relu blocks and then a pooling layer
    to halve the feature size"""
    out = nn.HybridSequential()
    out.add(nn.Conv2D(num_filters, 1, strides=1, padding=0))

    if norm_layer == nn.BatchNorm:
        out.add(nn.BatchNorm(in_channels=num_filters))
    elif norm_layer == contrib_nn.SyncBatchNorm:
        out.add(mx.gluon.contrib.nn.SyncBatchNorm(in_channels=num_filters, num_devices=norm_kwargs['num_devices']))
    else:
        raise ValueError("Unknown norm layer type: {}".format(norm_layer))
    if use_mish:
        out.add(Mish())
    else:
        out.add(nn.LeakyReLU(alpha=0.25))
    out.add(nn.Conv2D(num_filters, 3, strides=1, padding=1))
    if norm_layer == nn.BatchNorm:
        out.add(nn.BatchNorm(in_channels=num_filters))
    elif norm_layer == contrib_nn.SyncBatchNorm:
        out.add(mx.gluon.contrib.nn.SyncBatchNorm(in_channels=num_filters, num_devices=norm_kwargs['num_devices']))
    else:
        raise ValueError("Unknown norm layer type: {}".format(norm_layer))
    if use_mish:
        out.add(Mish())
    else:
        out.add(nn.LeakyReLU(alpha=0.25))
    out.add(nn.Conv2D(num_filters, 3, strides=stride, padding=1))
    if norm_layer == nn.BatchNorm:
        out.add(nn.BatchNorm(in_channels=num_filters))
    elif norm_layer == contrib_nn.SyncBatchNorm:
        out.add(mx.gluon.contrib.nn.SyncBatchNorm(in_channels=num_filters, num_devices=norm_kwargs['num_devices']))
    else:
        raise ValueError("Unknown norm layer type: {}".format(norm_layer))
    if use_mish:
        out.add(Mish())
    else:
        out.add(nn.LeakyReLU(alpha=0.25))
    return out


def convolution(num_filters, filter_size, norm_layer, norm_kwargs, stride=1, pad=1, use_mish=False):
    out = nn.HybridSequential()
    out.add(nn.Conv2D(num_filters, filter_size, padding=pad, strides=stride))
    if norm_layer == nn.BatchNorm:
        out.add(nn.BatchNorm(in_channels=num_filters))
    elif norm_layer == contrib_nn.SyncBatchNorm:
        out.add(mx.gluon.contrib.nn.SyncBatchNorm(in_channels=num_filters, num_devices=norm_kwargs['num_devices']))
    else:
        raise ValueError("Unknown norm layer type: {}".format(norm_layer))
    if use_mish:
        out.add(Mish())
    else:
        out.add(nn.LeakyReLU(alpha=0.25))
    return out


def deform_convolution(num_filters, norm_layer, norm_kwargs, use_mish=False):
    out = nn.HybridSequential()
    out.add(mx.gluon.contrib.cnn.DeformableConvolution(num_filters, kernel_size=(3, 3),
                                                       padding=(1, 1),
                                                       num_deformable_group=4,
                                                       weight_initializer=mx.init.Xavier(rnd_type='gaussian',
                                                                                         factor_type='out',
                                                                                         magnitude=2),
                                                       offset_weight_initializer=mx.init.Xavier(rnd_type='gaussian',
                                                                                                factor_type='out',
                                                                                                magnitude=2)))

    if norm_layer == nn.BatchNorm:
        out.add(nn.BatchNorm(in_channels=num_filters))
    elif norm_layer == contrib_nn.SyncBatchNorm:
        out.add(mx.gluon.contrib.nn.SyncBatchNorm(in_channels=num_filters, num_devices=norm_kwargs['num_devices']))
    else:
        raise ValueError("Unknown norm layer type: {}".format(norm_layer))
    if use_mish:
        out.add(Mish())
    else:
        out.add(nn.LeakyReLU(alpha=0.25))
    return out



"""SSD default config"""
# pylint: disable=unused-variable,missing-function-docstring,bad-whitespace,missing-class-docstring
from typing import Union, Tuple
from autocfg import dataclass, field


@dataclass
class SSD:
    # Backbone network.
    backbone : str = 'vgg16_atrous'  # base feature network
    # Input data shape, use 300, 512.
    data_shape : int = 300
    # List of convolution layer channels which is going to be appended to the base
    # network feature extractor. If `name` is `None`, this is ignored.
    filters : Union[int, None] = None
    # Sizes of anchor boxes, this should be a list of floats, in incremental order.
    # The length of `sizes` must be len(layers) + 1.
    sizes : Union[Tuple, list] = (30, 60, 111, 162, 213, 264, 315)
    # Aspect ratios of anchors in each output layer. Its length must be equals
    # to the number of SSD output layers.
    ratios : Union[Tuple, list] = tuple([[1, 2, 0.5]] + [[1, 2, 0.5, 3, 1.0/3]] * 3 + [[1, 2, 0.5]] * 2)
    # Step size of anchor boxes in each output layer.
    steps : Union[Tuple, list] = (8, 16, 32, 64, 100, 300)
    # Use synchronize BN across devices.
    syncbn : bool = False
    # Whether to use automatic mixed precision
    amp : bool = False
    # Whether to enable custom model.
    # custom_model = True
    # whether apply transfer learning from pre-trained models, if True, override other net structures
    transfer : Union[str, None] = 'ssd_512_resnet50_v1_coco'


@dataclass
class TrainCfg:
    # Batch size during training
    batch_size : int = 16
    # starting epoch
    start_epoch : int = 0
    # total epoch for training
    epochs : int = 20
    # Learning rate.
    lr : float = 0.001
    # Decay rate of learning rate.
    lr_decay : float = 0.1
    # Epochs at which learning rate decays
    lr_decay_epoch : Union[Tuple, list] = (160, 200)
    # Momentum
    momentum : float = 0.9
    # Weight decay
    wd : float = 5e-4
    # log interval in terms of iterations
    log_interval : int = 100
    # Random seed to be fixed.
    seed : int = 233
    # Use DALI for data loading and data preprocessing in training.
    # Currently supports only COCO.
    dali : bool = False


@dataclass
class ValidCfg:
    # Batch size during training
    batch_size : int = 16
    # Epoch interval for validation
    val_interval : int = 1
    # metric, 'voc', 'voc07'
    metric : str = 'voc07'
    # iou_thresh for VOC type metrics
    iou_thresh : float = 0.5

@dataclass
class SSDCfg:
    ssd : SSD = field(default_factory=SSD)
    train : TrainCfg = field(default_factory=TrainCfg)
    valid : ValidCfg = field(default_factory=ValidCfg)
    # Dataset name. eg. 'coco', 'voc', 'voc_tiny'
    dataset : str = 'voc_tiny'
    # Path of the directory where the dataset is located.
    dataset_root : str = '~/.mxnet/datasets/'
    # Training with GPUs, you can specify (1,3) for example.
    gpus : Union[Tuple, list] = (0, 1, 2, 3)
    # Resume from previously saved parameters if not None.
    # For example, you can resume from ./faster_rcnn_xxx_0123.params.
    resume : str = ''
    # Saving parameters epoch interval, best model will always be saved.
    save_interval : int = 1
    # Use MXNet Horovod for distributed training. Must be run with OpenMPI.
    horovod : bool = False
    # Number of data workers, you can use larger number to accelerate data loading,
    # if your CPU and GPUs are powerful.
    num_workers : int = 4



#!/usr/bin/env python
import os
from datetime import datetime
import io
import re
import shutil
import sys
from setuptools import setup, find_packages


def read(*names, **kwargs):
    with io.open(
        os.path.join(os.path.dirname(__file__), *names),
        encoding=kwargs.get("encoding", "utf8")
    ) as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")

long_description = (
"""GluonCV Python Package
=========================
`GluonCV <https://gluon-cv.mxnet.io>`_ provides implementations of the state-of-the-art (SOTA) deep learning models in computer vision.

It is designed for engineers, researchers, and students to fast prototype products and research ideas based on these models.

Installation
------------

To install, use:

.. code-block:: bash

    pip install gluoncv mxnet-mkl>=1.4.0 --upgrade

To enable different hardware supports such as GPUs, check out  `mxnet variants <https://pypi.org/project/mxnet/>`_.

For example, you can install cuda-9.0 supported mxnet alongside gluoncv:

.. code-block:: bash

    pip install gluoncv mxnet-cu90mkl>=1.4.0 --upgrade

""")

VERSION = find_version('gluoncv', '__init__.py')

if 'TRAVIS_TAG' in os.environ and os.environ['TRAVIS_TAG'].startswith('patch-'):
    VERSION = os.environ['TRAVIS_TAG'].split('-')[1]
elif 'APPVEYOR_REPO_TAG_NAME' in os.environ and os.environ['APPVEYOR_REPO_TAG_NAME'].startswith('patch-'):
    VERSION = os.environ['APPVEYOR_REPO_TAG_NAME'].split('-')[1]
elif 'TRAVIS_TAG' in os.environ or 'APPVEYOR_REPO_TAG_NAME' in os.environ:
    pass
else:
    VERSION += 'b{0}'.format(datetime.today().strftime('%Y%m%d'))

requirements = [
    'numpy',
    'tqdm',
    'requests',
    # 'mxnet',
    'matplotlib',
    'portalocker',
    'Pillow',
    'scipy',
    'tensorboardx',
    'decord',
    'opencv-python',
    'yacs',
    'pandas',
    'pyyaml',
    'autocfg',
    'autogluon.core'
]

setup(
    # Metadata
    name='gluoncv',
    version=VERSION,
    author='Gluon CV Toolkit Contributors',
    url='https://github.com/dmlc/gluon-cv',
    description='MXNet Gluon CV Toolkit',
    long_description=long_description,
    license='Apache-2.0',

    # Package info
    packages=find_packages(exclude=('docs', 'tests', 'scripts')),
    zip_safe=True,
    include_package_data=True,
    install_requires=requirements,
)



"""Auto pipeline for object detection task"""
# pylint: disable=bad-whitespace,missing-class-docstring
import time
import copy
import logging
import pickle
import pprint
from typing import Union, Tuple

from autocfg import dataclass
import numpy as np
import pandas as pd
import autogluon.core as ag
from autogluon.core.decorator import sample_config
from autogluon.core.scheduler.resource import get_cpu_count, get_gpu_count
from autogluon.core.task.base import BaseTask
from autogluon.core.searcher import RandomSearcher

from ..estimators.base_estimator import BaseEstimator
from ..estimators import SSDEstimator, FasterRCNNEstimator, YOLOv3Estimator, CenterNetEstimator
from .utils import auto_suggest, config_to_nested
from .dataset import ObjectDetectionDataset

__all__ = ['ObjectDetection']

@dataclass
class LiteConfig:
    transfer : Union[str, ag.Space, type(None)] = ag.Categorical('ssd_512_mobilenet1.0_coco', 'yolo3_mobilenet1.0_coco')
    lr : Union[ag.Space, float] = 1e-3
    num_trials : int = 1
    epochs : Union[ag.Space, int] = 5
    nthreads_per_trial : int = 32
    ngpus_per_trial : int = 0
    time_limits : int = 3600
    search_strategy : Union[str, ag.Space] = 'random'
    dist_ip_addrs : Union[type(None), list, Tuple] = None

@dataclass
class DefaultConfig:
    transfer : Union[ag.Space, str, type(None)] = ag.Categorical('yolo3_darknet53_coco', 'ssd_512_resnet50_v1_voc')
    lr : Union[ag.Space, float] = ag.Categorical(1e-3, 5e-3)
    num_trials : int = 3
    epochs : Union[ag.Space, int] = 10
    nthreads_per_trial : int = 128
    ngpus_per_trial : int = 8
    time_limits : int = 3600
    search_strategy : Union[str, ag.Space] = 'random'
    dist_ip_addrs : Union[type(None), list, Tuple] = None


@ag.args()
def _train_object_detection(args, reporter):
    """
    Parameters
    ----------
    args: <class 'autogluon.utils.edict.EasyDict'>
    """
    # train, val data
    train_data = args.pop('train_data')
    val_data = args.pop('val_data')
    # convert user defined config to nested form
    args = config_to_nested(args)

    tic = time.time()
    try:
        estimator_cls = args.pop('estimator', None)
        estimator = estimator_cls(args, reporter=reporter)
        # training
        result = estimator.fit(train_data=train_data, val_data=val_data)
    # pylint: disable=bare-except
    except:
        import traceback
        return {'traceback': traceback.format_exc(), 'args': str(args),
                'time': time.time() - tic, 'train_map': -1, 'valid_map': -1}

    # TODO: checkpointing needs to be done in a better way
    # unique_checkpoint = 'train_object_detection_' + str(uuid.uuid4()) + '.pkl'
    # estimator.save(unique_checkpoint)
    result.update({'model_checkpoint': pickle.dumps(estimator)})
    return result

class ObjectDetection(BaseTask):
    """Object Detection general task.

    Parameters
    ----------
    config : dict
        The configurations, can be nested dict.
    logger : logging.Logger
        The desired logger object, use `None` for module specific logger with default setting.

    """
    Dataset = ObjectDetectionDataset

    def __init__(self, config=None, logger=None):
        super(ObjectDetection, self).__init__()
        self._fit_summary = {}
        self._logger = logger if logger is not None else logging.getLogger(__name__)
        self._logger.setLevel(logging.INFO)

        # cpu and gpu setting
        cpu_count = get_cpu_count()
        gpu_count = get_gpu_count()

        # default settings
        if not config:
            if gpu_count < 1:
                self._logger.info('No GPU detected/allowed, using most conservative search space.')
                config = LiteConfig()
            else:
                config = DefaultConfig()
            config = config.asdict()
        else:
            if not config.get('dist_ip_addrs', None):
                ngpus_per_trial = config.get('ngpus_per_trial', gpu_count)
                if ngpus_per_trial < 1:
                    self._logger.info('No GPU detected/allowed, using most conservative search space.')
                    default_config = LiteConfig()
                else:
                    default_config = DefaultConfig()
                config = default_config.merge(config, allow_new_key=True).asdict()

        # adjust cpu/gpu resources
        if not config.get('dist_ip_addrs', None):
            nthreads_per_trial = config.get('nthreads_per_trial', cpu_count)
            if nthreads_per_trial > cpu_count:
                nthreads_per_trial = cpu_count
            ngpus_per_trial = config.get('ngpus_per_trial', gpu_count)
            if ngpus_per_trial > gpu_count:
                ngpus_per_trial = gpu_count
                self._logger.warning(
                    "The number of requested GPUs is greater than the number of available GPUs."
                    "Reduce the number to %d", ngpus_per_trial)
        else:
            raise ValueError('Please specify `nthreads_per_trial` and `ngpus_per_trial` '
                             'given that dist workers are available')

        # fix estimator-transfer relationship
        estimator = config.get('estimator', None)
        transfer = config.get('transfer', None)
        if estimator is not None and transfer is not None:
            if isinstance(transfer, ag.Space):
                transfer = transfer.data
            if isinstance(transfer, str):
                transfer = [transfer]
            transfer = [t for t in transfer if estimator in t]
            if not transfer:
                raise ValueError(f'No matching `transfer` model for {estimator}')
            if len(transfer) == 1:
                config['transfer'] = transfer[0]
            else:
                config['transfer'] = ag.Categorical(**transfer)

        # additional configs
        config['num_workers'] = nthreads_per_trial
        config['gpus'] = [int(i) for i in range(ngpus_per_trial)]
        if config['gpus']:
            config['batch_size'] = config.get('batch_size', 8) * len(config['gpus'])
            self._logger.info('Increase batch size to %d based on the number of gpus %d',
                              config['batch_size'], len(config['gpus']))
        config['seed'] = config.get('seed', np.random.randint(32,767))
        self._config = config

        # scheduler options
        self.search_strategy = config.get('search_strategy', 'random')
        self.scheduler_options = {
            'resource': {'num_cpus': nthreads_per_trial, 'num_gpus': ngpus_per_trial},
            'checkpoint': config.get('checkpoint', 'checkpoint/exp1.ag'),
            'num_trials': config.get('num_trials', 2),
            'time_out': config.get('time_limits', 60 * 60),
            'resume': (len(config.get('resume', '')) > 0),
            'visualizer': config.get('visualizer', 'none'),
            'time_attr': 'epoch',
            'reward_attr': 'map_reward',
            'dist_ip_addrs': config.get('dist_ip_addrs', None),
            'searcher': self.search_strategy,
            'search_options': config.get('search_options', None)}
        if self.search_strategy == 'hyperband':
            self.scheduler_options.update({
                'searcher': 'random',
                'max_t': config.get('epochs', 50),
                'grace_period': config.get('grace_period', config.get('epochs', 50) // 4)})

    def fit(self, train_data, val_data=None, train_size=0.9, random_state=None):
        """Fit auto estimator given the input data.

        Parameters
        ----------
        train_data : pd.DataFrame or iterator
            Training data.
        val_data : pd.DataFrame or iterator, optional
            Validation data, optional. If `train_data` is DataFrame, `val_data` will be split from
            `train_data` given `train_size`.
        train_size : float
            The portion of train data split from original `train_data` if `val_data` is not provided.
        random_state : int
            Random state for splitting, for `np.random.seed`.

        Returns
        -------
        Estimator
            The estimator obtained by training on the specified dataset.

        """
        # split train/val before HPO to make fair comparisons
        if not isinstance(train_data, pd.DataFrame):
            assert val_data is not None, \
                "Please provide `val_data` as we do not know how to split `train_data` of type: \
                {}".format(type(train_data))

        if val_data is None:
            assert 0 <= train_size <= 1.0
            if random_state:
                np.random.seed(random_state)
            split_mask = np.random.rand(len(train_data)) < train_size
            train = train_data[split_mask]
            val = train_data[~split_mask]
            self._logger.info('Randomly split train_data into train[%d]/validation[%d] splits.',
                              len(train), len(val))
            train_data, val_data = train, val

        # automatically suggest some hyperparameters based on the dataset statistics(experimental)
        if not self._config.get('transfer', None):
            estimator = self._config.get('estimator', None)
            if estimator is None:
                estimator = [SSDEstimator, FasterRCNNEstimator, YOLOv3Estimator, CenterNetEstimator]
            self._config['estimator'] = ag.Categorical(*estimator)
            self._config['train_dataset'] = train_data
            if self._config.get('auto_suggest', True):
                auto_suggest(self._config, estimator, self._logger)
            self._config.pop('train_dataset')

        # register args
        config = self._config.copy()
        config['train_data'] = train_data
        config['val_data'] = val_data
        _train_object_detection.register_args(**config)

        start_time = time.time()
        self._fit_summary = {}
        if config.get('num_trials', 1) < 2:
            rand_config = RandomSearcher(_train_object_detection.cs).get_config()
            self._logger.info("Starting fit without HPO")
            results = _train_object_detection(_train_object_detection.args, rand_config)
            best_config = sample_config(_train_object_detection.args, rand_config)
            best_config.pop('train_data', None)
            best_config.pop('val_data', None)
            self._fit_summary.update({'train_map': results.get('train_map', -1),
                                      'valid_map': results.get('valid_map', -1),
                                      'total_time': results.get('time', time.time() - start_time),
                                      'best_config': best_config})
        else:
            self._logger.info("Starting HPO experiments")
            results = self.run_fit(_train_object_detection, self.search_strategy,
                                   self.scheduler_options)
        end_time = time.time()
        self._logger.info(">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> finish model fitting")
        self._logger.info("total runtime is %.2f s", end_time - start_time)
        if config.get('num_trials', 1) > 1:
            best_config = sample_config(_train_object_detection.args, results['best_config'])
            # convert best config to nested form
            best_config = config_to_nested(best_config)
            best_config.pop('train_data', None)
            best_config.pop('val_data', None)
            self._fit_summary.update({'train_map': results.get('train_map', -1),
                                      'valid_map': results.get('valid_map', results.get('best_reward', -1)),
                                      'total_time': results.get('total_time', time.time() - start_time),
                                      'best_config': best_config})
        self._logger.info(pprint.pformat(self._fit_summary, indent=2))

        # TODO: checkpointing needs to be done in a better way
        model_checkpoint = results.get('model_checkpoint', None)
        if model_checkpoint is None:
            raise RuntimeError(f'Unexpected error happened during fit: {pprint.pformat(results, indent=2)}')
        estimator = pickle.loads(results['model_checkpoint'])
        return estimator

    def fit_summary(self):
        return copy.copy(self._fit_summary)

    @classmethod
    def load(cls, filename):
        obj = BaseEstimator.load(filename)
        # make sure not accidentally loading e.g. classification model
        # pylint: disable=unidiomatic-typecheck
        assert type(obj) in (SSDEstimator, FasterRCNNEstimator, YOLOv3Estimator, CenterNetEstimator)
        return obj



# pylint: disable=wildcard-import, unused-wildcard-import
"""
GluonCV-PyTorch model zoo
"""
from .action_recognition import *


__all__ = ['get_model', 'get_model_list']


_models = {
    'resnet18_v1b_kinetics400': resnet18_v1b_kinetics400,
    'resnet34_v1b_kinetics400': resnet34_v1b_kinetics400,
    'resnet50_v1b_kinetics400': resnet50_v1b_kinetics400,
    'resnet101_v1b_kinetics400': resnet101_v1b_kinetics400,
    'resnet152_v1b_kinetics400': resnet152_v1b_kinetics400,
    'resnet50_v1b_sthsthv2': resnet50_v1b_sthsthv2,
    'resnet50_v1b_custom': resnet50_v1b_custom,
    'i3d_resnet50_v1_kinetics400': i3d_resnet50_v1_kinetics400,
    'i3d_resnet101_v1_kinetics400': i3d_resnet101_v1_kinetics400,
    'i3d_nl5_resnet50_v1_kinetics400': i3d_nl5_resnet50_v1_kinetics400,
    'i3d_nl10_resnet50_v1_kinetics400':i3d_nl10_resnet50_v1_kinetics400,
    'i3d_nl5_resnet101_v1_kinetics400': i3d_nl5_resnet101_v1_kinetics400,
    'i3d_nl10_resnet101_v1_kinetics400': i3d_nl10_resnet101_v1_kinetics400,
    'i3d_resnet50_v1_sthsthv2': i3d_resnet50_v1_sthsthv2,
    'i3d_resnet50_v1_custom': i3d_resnet50_v1_custom,
    'slowfast_4x16_resnet50_kinetics400': slowfast_4x16_resnet50_kinetics400,
    'slowfast_8x8_resnet50_kinetics400': slowfast_8x8_resnet50_kinetics400,
    'slowfast_4x16_resnet101_kinetics400': slowfast_4x16_resnet101_kinetics400,
    'slowfast_8x8_resnet101_kinetics400': slowfast_8x8_resnet101_kinetics400,
    'slowfast_16x8_resnet101_kinetics400': slowfast_16x8_resnet101_kinetics400,
    'slowfast_16x8_resnet101_50_50_kinetics400': slowfast_16x8_resnet101_50_50_kinetics400,
    'slowfast_16x8_resnet50_sthsthv2': slowfast_16x8_resnet50_sthsthv2,
    'slowfast_4x16_resnet50_custom': slowfast_4x16_resnet50_custom,
    'i3d_slow_resnet50_f32s2_kinetics400': i3d_slow_resnet50_f32s2_kinetics400,
    'i3d_slow_resnet50_f16s4_kinetics400': i3d_slow_resnet50_f16s4_kinetics400,
    'i3d_slow_resnet50_f8s8_kinetics400': i3d_slow_resnet50_f8s8_kinetics400,
    'i3d_slow_resnet101_f32s2_kinetics400': i3d_slow_resnet101_f32s2_kinetics400,
    'i3d_slow_resnet101_f16s4_kinetics400': i3d_slow_resnet101_f16s4_kinetics400,
    'i3d_slow_resnet101_f16s4_kinetics700': i3d_slow_resnet101_f16s4_kinetics700,
    'i3d_slow_resnet101_f8s8_kinetics400': i3d_slow_resnet101_f8s8_kinetics400,
    'i3d_slow_resnet50_f32s2_custom': i3d_slow_resnet50_f32s2_custom,
    'r2plus1d_v1_resnet18_kinetics400': r2plus1d_v1_resnet18_kinetics400,
    'r2plus1d_v1_resnet34_kinetics400': r2plus1d_v1_resnet34_kinetics400,
    'r2plus1d_v1_resnet50_kinetics400': r2plus1d_v1_resnet50_kinetics400,
    'r2plus1d_v1_resnet101_kinetics400': r2plus1d_v1_resnet101_kinetics400,
    'r2plus1d_v1_resnet152_kinetics400': r2plus1d_v1_resnet152_kinetics400,
    'r2plus1d_v2_resnet152_kinetics400': r2plus1d_v2_resnet152_kinetics400,
    'r2plus1d_v1_resnet50_custom': r2plus1d_v1_resnet50_custom,
    'ircsn_v2_resnet152_f32s2_kinetics400': ircsn_v2_resnet152_f32s2_kinetics400,
    'tpn_resnet50_f8s8_kinetics400': tpn_resnet50_f8s8_kinetics400,
    'tpn_resnet50_f16s4_kinetics400': tpn_resnet50_f16s4_kinetics400,
    'tpn_resnet50_f32s2_kinetics400': tpn_resnet50_f32s2_kinetics400,
    'tpn_resnet101_f8s8_kinetics400': tpn_resnet101_f8s8_kinetics400,
    'tpn_resnet101_f16s4_kinetics400': tpn_resnet101_f16s4_kinetics400,
    'tpn_resnet101_f32s2_kinetics400': tpn_resnet101_f32s2_kinetics400,
    'tpn_resnet50_f32s2_custom': tpn_resnet50_f32s2_custom,
}


def get_model(cfg):
    """Returns a pre-defined model by name

    Returns
    -------
    The model.
    """
    name = cfg.CONFIG.MODEL.NAME.lower()
    if name not in _models:
        err_str = '"%s" is not among the following model list:\n\t' % (name)
        err_str += '%s' % ('\n\t'.join(sorted(_models.keys())))
        raise ValueError(err_str)
    net = _models[name](cfg)
    return net


def get_model_list():
    """Get the entire list of model names in model_zoo.

    Returns
    -------
    list of str
        Entire list of model names in model_zoo.

    """
    return _models.keys()
